#include "arrange.h"

#ifdef WIN32
#define nextafter _nextafter
#endif

Segment::Segment (const Point &t, const Point &h, bool iin, Splat *isplat)
  : circle(false), np(2), ip(0), points(new Point [2]), edge(0),
    splat(isplat)
{
  if (t < h) {
    points[0] = t;
    points[1] = h;
    in = iin;
  }
  else {
    points[0] = h;
    points[1] = t;
    in = !iin;
  }
}

Segment::Segment (const Point &t, const Point &h, const Point &io,
		  double ir, bool iupper, bool iin)
  : circle(t.x != h.x), upper(iupper), np(2), ip(0), points(new Point [2]), 
    o(io), r(ir), edge(0), splat(0)
{
  if (t < h) {
    points[0] = t;
    points[1] = h;
    in = iin;
  }
  else {
    points[0] = h;
    points[1] = t;
    in = !iin;
  }
}

Segment::Segment (const Points &ipoints, bool iin)
  : circle(false), np(ipoints.size()), ip(0), points(new Point [np]), 
    edge(0), splat(0)
{
  if (ipoints[0] < ipoints[np-1]) {
    for (int i = 0; i < np; ++i)
      points[i] = ipoints[i];
    in = iin;
  }
  else {
    for (int i = 0; i < np; ++i)
      points[i] = ipoints[np-i-1];
    in = !iin;
  }
}

double Segment::ymax () const
{
  if (circle && upper && points[0].x < o.x && points[1].x > o.x)
    return o.y + r;
  return max(points[ip].y, points[ip+1].y);
}

double Segment::ymin () const
{
  if (circle && !upper && points[0].x < o.x && points[1].x > o.x)
    return o.y - r;
  return min(points[ip].y, points[ip+1].y);
}

double Segment::y (double x, bool ipflag) const
{
  if (circle) {
    double d = r*r - (x - o.x)*(x - o.x);
    d = d < 0.0 ? 0.0 : sqrt(d);
    return upper ? o.y + d : o.y - d;
  }
  int i;
  if (!ipflag) {
    i = 0;
    while (i < np - 2 && points[i+1].x <= x)
      ++i;
  }
  else 
    i = ip;    
  return line_y(x, points[i], points[i+1]);
}

double line_y (double x, const Point &p, const Point &q)
{
  double u = (x - p.x)/(q.x - p.x);
  return (1.0 - u)*p.y + u*q.y;
}

double Segment::dy (double x) const
{
  if (circle) {
    double d = r*r - (x - o.x)*(x - o.x);
    d = d < 0.0 ? 0.0 : sqrt(d);
    if (d == 0.0) return 1e20;
    return upper ? (o.x - x)/d : (x - o.x)/d;
  }
  return (points[ip+1].y - points[ip].y)/(points[ip+1].x - points[ip].x);
}

Rootlist Segment::rootlist (const Segment &s) const
{
  bool flag = this < &s;
  const Segment *a = flag ? this : &s, *b = flag ? &s : this;
  Rootlist rl = (circle || s.circle) ? a->rootlist_g(*b) : a->rootlist_ll(*b);
  if (!flag)
    rl.below = !rl.below;
  return rl;
}

Rootlist Segment::rootlist_g (const Segment &s) const
{
  if (ymax() <= s.ymin())
    return Rootlist(true);
  if (s.ymax() <= ymin())
    return Rootlist(false);
  double rlb, rub;
  if (circle && s.circle) {
    rlb = max(o.x - r, s.o.x - s.r);
    rub = min(o.x + r, s.o.x + s.r);
  }
  else if (circle) {
    rlb = o.x - r;
    rub = o.x + r;
  }
  else {
    rlb = s.o.x - s.r;
    rub = s.o.x + s.r;
  }
  double xs = rlb, w = 0.0, xm, x[2];
  double x0 = max(lb(), s.lb()), xn = min(ub(), s.ub());
  int n = intersection(s, x);
  int iw, i0 = 0;
  for (int i = 0; i <= n; ++i) {
    double xe = i == n ? rub : x[i];
    if (xe - xs > w) {
      w = xe - xs;
      xm = 0.5*(xs + xe);
      iw = i;
    }
    if (xs <= x0 && x0 < xe)
      i0 = i;
    xs = xe;
  }
  double dy = s.y(xm) - y(xm);
  bool below = (i0 - iw)%2 == 0 ? dy > 0.0 : dy < 0.0;
  Rootlist rl(below);

  for (int i = 0; i < n; ++i)
    if (x[i] > x0 && x[i] < xn) {
      rl.x[rl.n] = x[i];
      ++rl.n;
    }
  return rl;
}

Rootlist Segment::rootlist_ll (const Segment &s) const
{
  const Point &ta(points[ip]), &ha(points[ip+1]),
    &tb = s.points[s.ip], &hb = s.points[s.ip+1];
  if (max(ta.y, ha.y) < min(tb.y, hb.y))
    return Rootlist(true);
  if (max(tb.y, hb.y) < min(ta.y, ha.y))
    return Rootlist(false);
  Point ua = ha - ta, ub = hb - tb;
  double sta = ub*(ta - tb), sha = ub*(ha - tb),
    stb = ua*(tb - ta), shb = ua*(hb - ta);
  if (sta <= 0.0 && sha <= 0.0 || stb >= 0.0 && shb >= 0.0)
    return Rootlist(true);
  if (sta >= 0.0 && sha >= 0.0 || stb <= 0.0 && shb <= 0.0)
    return Rootlist(false);
  bool below = sha > sta;
  double rlb = max(ta.x, tb.x), rub = min(ha.x, hb.x);
  if (rlb < rub && nextafter(rlb, rub) == rub)
    return Rootlist(below);
  Rootlist rl(below);
  rl.n = 1;
  int n = intersection(s, rl.x);
  if (n == 0 || rl.x[0] >= rub)
    rl.x[0] = nextafter(rub, rlb);
  else if (rl.x[0] <= rlb)
    rl.x[0] = nextafter(rlb, rub);
  return rl;
}

int Segment::intersection (const Segment &s, double *x) const
{
  if (circle)
    return s.circle ? intersection_cc(s, x) : intersection_cl(s, x);
  else if (s.circle)
    return s.intersection_cl(*this, x);
  else
    return intersection_ll(s, x);
}

int Segment::intersection_ll (const Segment &s, double *x) const
{
  const Point &t = points[ip], &h = points[ip+1], &st = s.points[s.ip],
    &sh = s.points[s.ip+1];
  if (t == st) {
    *x = t.x;
    return 1;
  }
  if (h == sh) {
    *x = h.x;
    return 1;
  }
  double a1, b1, c1, a2, b2, c2, xy[2];
  a1 = t.y - h.y;
  b1 = h.x - t.x;
  c1 = - a1*h.x - b1*h.y;
  a2 = st.y - sh.y;
  b2 = sh.x - st.x;
  c2 = - a2*sh.x - b2*sh.y;
  if (!line_intersection(a1, b1, c1, a2, b2, c2, xy))
    return 0;
  *x = xy[0];
  return 1;
}

bool line_intersection (double a1, double b1, double c1, double a2,
			double b2, double c2, double *res)
{
  double delta = a1*b2 - a2*b1;
  if (delta == 0.0)
    return false;
  res[0] = (c2*b1 - c1*b2)/delta;
  res[1] = (c1*a2 - c2*a1)/delta;
  return true;
}

int Segment::intersection_cl (const Segment &s, double *x) const
{
  const Point &st = s.points[s.ip], &sh = s.points[s.ip+1];
  double a, b, c, xx[4];
  a = st.y - sh.y;
  b = sh.x - st.x;
  c = - a*sh.x - b*sh.y;
  if (!circle_line_intersections(o.x, o.y, r, a, b, c, xx))
    return 0;
  int m = 0;
  for (int i = 0; i < 2; ++i)
    if (upper ? xx[2*i+1] > o.y : xx[2*i+1] < o.y)
      x[m++] = xx[2*i];
  return m;
}

bool circle_line_intersections (double ox, double oy, double r,
				double a, double b, double c, double *res)
{
  double den = 1.0/sqrt(a*a + b*b);
  a *= den;
  b *= den;
  c *= den;
  double d = a*ox + b*oy + c;
  if (fabs(d) >= r)
    return false;
  double e = sqrt(r*r - d*d), px = ox - d*a, py = oy - d*b;
  res[0] = px - e*b;
  res[1] = py + e*a;
  res[2] = px + e*b;
  res[3] = py - e*a;
  return true;
}

int Segment::intersection_cc (const Segment &s, double *x) const
{
  double xy[4];
  if (!circle_intersections(o.x, o.y, r, s.o.x, s.o.y, s.r, xy))
    return 0;
  int m = 0;
  for (int i = 0; i < 2; ++i) {
    bool fa = upper ? xy[2*i+1] > o.y : xy[2*i+1] < o.y;
    bool fb = s.upper ? xy[2*i+1] > s.o.y : xy[2*i+1] < s.o.y;
    if (fa && fb)
      x[m++] = xy[2*i];
  }
  if (m == 2 && x[0] > x[1]) {
    double t = x[0];
    x[0] = x[1];
    x[1] = t;
  }
  return m;
}

bool circle_intersections (double x1, double y1, double r1, 
			   double x2, double y2, double r2, 
			   double *res, bool safe)
{ 
  double delta = safe ? 1e-12 : 0.0;
  double d = (y2 - y1)*(y2 - y1) + (x2 - x1)*(x2 - x1);
  if (d < (r1 - r2)*(r1 - r2) - delta || 
      d > (r1 + r2)*(r1 + r2) + delta)
    return false;
  d = sqrt(d);
  double ct = (r1*r1 + d*d - r2*r2)/(2.0*r1*d);
  double st = 1.0 - ct*ct;
  st = st < 0.0 ? 0.0 : sqrt(st);
  double u = r1*(x2 - x1)/d, v = r1*(y2 - y1)/d;
  res[0] = x1 + ct*u - st*v;
  res[1] = y1 + st*u + ct*v;
  res[2] = x1 + ct*u + st*v;
  res[3] = y1 - st*u + ct*v;
  return true;
}

bool Segment::below (Segment *s, double x) const
{
  Rootlist rl = rootlist(*s);
  bool res = rl.below;
  for (int i = 0; i < rl.n && x >= rl.x[i]; ++i)
    res = !res;
  return res;
}
 
bool Segment::next (Segment *s, double x, double &nx) const
{
  Rootlist rl = rootlist(*s);
  for (int i = 0; i < rl.n; ++i)
    if (x < rl.x[i]) {
      nx = rl.x[i];
      return true;
    }
  return false;
}

void Segment::bounding_box (double xs, double xe, double *bbox) const
{
  if (circle) {
    if (o.x > xs && o.x < xe)
      if (upper)
	bbox[1] = o.y + r;
      else
	bbox[0] = o.y - r;
    return;
  }
  int i = 0;
  while (i <np && xs >= points[i].x)
    ++i;
  while (i < np && points[i].x < xe) {
    if (points[i].y < bbox[0])
      bbox[0] = points[i].y;
    else if (points[i].y > bbox[1])
      bbox[1] = points[i].y;
    ++i;
  }
}

void Segment::discretize (double xs, double xe, double tol, Points &res) const
{
  if (circle)
    discretize_circle(xs, xe, tol, res);
  else
    discretize_pline(xs, xe, res);
}

void Segment::discretize_circle (double xs, double xe, double tol, 
				 Points &res) const
{
  double th = 2.0*acos(1.0 - tol/r);
  if (upper) th = - th;
  double s = sin(th), c = cos(th), px = xs - o.x, py = y(xs) - o.y;
  res.push_back(Point(xs, y(xs)));
  while (1) {
    double nx = c*px - s*py, ny = s*px + c*py;
    if (o.x + nx >= xe || (upper ? ny <= 0.0 : ny >= 0.0))
      break;
    res.push_back(Point(o.x + nx, o.y + ny));
    px = nx;
    py = ny;
  }
  res.push_back(Point(xe, y(xe)));
}

void Segment::discretize_pline (double xs, double xe, Points &res) const
{
  if (vertical()) {
    res.push_back(points[0]);
    res.push_back(points[1]);
    return;
  }
  int i = 1;
  while (xs >= points[i].x)
    ++i;
  res.push_back(Point(xs, line_y(xs, points[i-1], points[i])));
  while (xe > points[i].x)
    res.push_back(points[i++]);
  res.push_back(Point(xe, line_y(xe, points[i-1], points[i])));
}

void delete_segments (Segments &segments)
{
  for (unsigned int i = 0; i < segments.size(); ++i)
    delete segments[i];
}

bool SegmentOrder::operator() (SegmentNode *a, SegmentNode *b) const
{ 
  double x = max(a->s->tail().x, b->s->tail().x);
  return a->s->below(b->s, x);
}

void SegmentTree::insert (Segment *s)
{
  s->node = data.insert(new SegmentNode(s)).first;
}

void SegmentTree::remove (Segment *s)
{
  SegmentNode *sn = *s->node;
  data.erase(s->node);
  delete sn;
}

void SegmentTree::swap (Segment *s, Segment *t)
{
  SegmentSet::iterator temp = s->node;
  s->node = t->node;
  t->node = temp;
  (*s->node)->s = s;
  (*t->node)->s = t;
}

Segment * SegmentTree::succ (Segment *s) const
{
  SegmentSet::iterator iter = s->node;
  ++iter;
  return iter == data.end() ? 0 : (*iter)->s;
}

Segment * SegmentTree::pred (Segment *s) const
{
  if (s->node == data.begin())
    return 0;
  SegmentSet::iterator iter = s->node;
  --iter;
  return (*iter)->s;
}

bool Event::operator< (const Event &b) const
{ 
  if (x > b.x)
    return true;
  if (x < b.x)
    return false;
  bool flag1 = (a->vertical() || type == Insert && !b.a->vertical()
		|| b.type == Remove || b.type == Shift);
  bool flag2 = (b.a->vertical() || b.type == Insert && !a->vertical()
		|| type == Remove || type == Shift);
  return flag1 && !flag2;
}

Eventheap::Eventheap (const Segments &segments)
{
  int m = segments.size();
  nmax = 10*m;
  events = new Event [nmax];
  n = 0;
  for (int i = 0; i < m; ++i)
    if (segments[i]->tail() != segments[i]->head()) {
      segments[i]->ip = 0;
      events[n].type = Insert;
      events[n].x = segments[i]->tail().x;
      events[n].a = segments[i];
      ++n;
    }
  make_heap(events, events + n);
}

inline Event * Eventheap::pop ()
{
  pop_heap(events, events + n);
  --n;
  return events + n;
}

inline void Eventheap::push (EventType type, double x, Segment *a,
			     Segment *b)
{
  if (n == nmax) {
    Event *nevents = new Event [2*n];
    for (int i = 0; i < n; ++i)
      nevents[i] = events[i];
    delete [] events;
    events = nevents;
    nmax *= 2;
  }
  Event *e = events + n;
  e->x = x;
  e->type = type;
  e->a = a;
  e->b = b;
  ++n;
  push_heap(events, events + n);
}

int Edge::set_heights (int h0)
{
  if (h > 0) 
    return h0;
  for (unsigned int i = 0; i < below.size(); ++i)
    h0 = below[i]->set_heights(h0);
  h = twin->h = h0;
  return h + 1;
}

int Edge::rank () const
{
  if (twin->tail.x < tail.x)
    return 0;
  else if (twin->tail.x > tail.x)
    return 2;
  else if (twin->tail.y > tail.y)
    return 1;
  else 
    return 3;
}

bool Edge::contains (const Point &p) const
{
  bool in = false;
  const Edge *e = this;
  do {
    if (e->above_point(p))
      in = !in;
    e = e->twin->next;
  }
  while (e != this);
  return in;
}

bool Edge::above_point (const Point &p) const
{
  const Point &a = tail, &b = twin->tail;
  if (p.x == a.x)
    return false;
  bool inc = b.x > a.x;
  if (inc ? p.x < a.x || p.x > b.x : p.x < b.x || p.x > a.x)
    return false;
  if (p.x == b.x) {
    const Point &c = twin->next->twin->tail;
    if (inc ? c.x < b.x : c.x > b.x)
      return false;
  }
  return segment->y(p.x, false) > p.y;
}

void Edge::discretize (double tol, Points &res) const
{
  if (segment->vertical() || tail.x == twin->tail.x) {
    res.push_back(tail);
    res.push_back(twin->tail);
    return;
  }
  double xt = tail.x, xh = twin->tail.x;
  if (inc())
    segment->discretize(xt, xh, tol, res);
  else {
    Points points;
    segment->discretize(xh, xt, tol, points);
    for (int i = int(points.size()) - 1; i >= 0; --i)
      res.push_back(points[i]);
  }
}

void Edge::restrict (Segments &res) const
{
  const Edge *e = this;
  do {
    const Point &te = e->tail, &he = e->twin->tail,
      &ts = te < he ? te : he, &hs = te < he ? he : te;
    Segment *s = e->segment;
    if (s->circle)
      res.push_back(new Segment(ts, hs, s->o, s->r, s->upper, s->in));
    else
      res.push_back(new Segment(ts, hs, s->in));
    e = e->twin->next;
  }
  while (e != this);
}

bool Edge::null_loop () const
{
  const Edge *e = this;
  do {
    if (e->in) return false;
    e = e->twin->next;
  }
  while (e != this);
  return true;
}

int Edge::loop_length () const
{
  const Edge *e = this;
  int n = 0;
  do {
    ++n;
    e = e->twin->next;
  }
  while (e != this);
  return n;
}

void Edge::bounding_box (double *bbox) const
{
  bbox[0] = min(tail.x, twin->tail.x);
  bbox[1] = max(tail.x, twin->tail.x);
  bbox[2] = min(tail.y, twin->tail.y);
  bbox[3] = max(tail.y, twin->tail.y);
  if (!segment->vertical())
    segment->bounding_box(bbox[0], bbox[1], bbox + 2);
}

void Face::form_loop (Edge *e)
{
  if (loops.empty())
    e->bounding_box(bbox);
  loops.push_back(e);
  for (Edge *f = e; !f->traversed; f = f->twin->next) {
    f->traversed = true;
    f->face = this;
    if (!f->tiny())
      in = in && f->in;
    double bboxf[4];
    f->bounding_box(bboxf);
    bbox[0] = min(bbox[0], bboxf[0]);
    bbox[1] = max(bbox[1], bboxf[1]);
    bbox[2] = min(bbox[2], bboxf[2]);
    bbox[3] = max(bbox[3], bboxf[3]);
  }
}

void Face::traverse (Profile &boundary) const
{
  for (unsigned int i = 0; i < loops.size(); ++i) {
    boundary.push_back(Points());
    Points &points = boundary[boundary.size() - 1];
    Edge *e = loops[i];
    do {
      points.push_back(e->tail);
      e = e->twin->next;
    }
    while (e != loops[i]);
  }
}

void Face::discretize (double tol, Profile &boundary, double dmin) const
{
  for (unsigned int i = 0; i < loops.size(); ++i) {
    Points points;
    Edge *e = loops[i];
    int n = 0;
    do {
      e->discretize(tol, points);
      e = e->twin->next;
      int nn = points.size();
      if (n > 0 && nn > n && points[n-1].x == points[n].x)
	points[n].y = points[n-1].y;
      n = nn;
    }
    while (e != loops[i]);
    points[n-1] = points[0];
    if (dmin == 0.0)
      boundary.push_back(points);
    else {
      boundary.push_back(Points());
      Points &loop = boundary[boundary.size() - 1];
      loop.push_back(points[0]);
      for (unsigned int i = 1; i < points.size(); ++i) {
	Point &p = loop[loop.size() - 1], q = points[i];
	if (fabs(p.x - q.x) > dmin || fabs(p.y - q.y) > dmin)
	  loop.push_back(q);
      }
    }
  }
}

bool Face::contains (const Point &p) const
{
  if (loops.empty() || !bbox_contains(bbox, p) || !loops[0]->contains(p))
    return false;
  for (unsigned int i = 1; i < loops.size(); ++i)
    if (loops[i]->contains(p))
      return false;
  return true;
}

bool bbox_contains (const double *bbox, const Point &p)
{
  return p.x >= bbox[0] && p.x <= bbox[1] && 
    p.y >= bbox[2] && p.y <= bbox[3];
}

Arrangement::~Arrangement ()
{
  for (unsigned int i = 0; i < edges.size(); ++i)
    delete edges[i];
  for (unsigned int i = 0; i < faces.size(); ++i)
    delete faces[i];
}

void Arrangement::arrange (Segments &segments)
{
  split_near_vertical(segments);
  sweep(segments);
  set_heights();
  embed();
  form_vertices();
  form_faces();
}

void Arrangement::split_near_vertical (Segments &segments) const
{
  int n = segments.size();
  for (int i = 0; i < n; ++i) {
    Segment *s = segments[i];
    if (s->np == 2) {
      double xt = s->points[0].x, xh = s->points[1].x;
      if (xt < xh && nextafter(xt, xh) == xh) {
	double yt = s->points[0].y, yh = s->points[1].y;
	if (fabs(yt - yh) > 1e-14) {
	  Point p(xh, yt), q(xh, yh);
	  s->points[1] = p;
	  segments.push_back(new Segment(p, q, s->in));
	}
      }
    }
  }
}

void Arrangement::sweep (const Segments &segments)
{
  SegmentTree sweepline;
  Eventheap events(segments);
  while (!events.empty()) {
    Event e = *events.pop();
    switch (e.type) {
    case Insert:
      insert(e.x, e.a, events, sweepline);
      break;
    case Shift:
      shift(e.x, e.a, events, sweepline);
      break;
    case Remove:
      remove(e.x, e.a, events, sweepline);
      break;
    case Swap:
      swap(e.x, e.a, e.b, events, sweepline);
      break;
    case Check:
      check(e.x, e.a, e.b, events, sweepline);
      break;
    }
  }
}

void Arrangement::insert (double x, Segment *a, Eventheap &events, 
			  SegmentTree &sweepline)
{
  if (a->vertical()) {
    forward_edge(a, a->tail());
    backward_edge(a, a->head());
    return;
  }
  sweepline.insert(a);
  Segment *pred = sweepline.pred(a), *succ = sweepline.succ(a);
  forward_edge(a, a->tail(), pred, succ);
  post_check(x, pred, a, events);
  post_check(x, a, succ, events);
  events.push(Shift, a->ub(), a);
}

void Arrangement::shift (double x, Segment *a, Eventheap &events, 
			 SegmentTree &sweepline)
{
  if (x < a->head().x) {
    a->shift();
    post_check(x, sweepline.pred(a), a, events);
    post_check(x, a, sweepline.succ(a), events);
    events.push(Shift, a->ub(), a);
  }
  else
    events.push(Remove, a->head().x, a);
}

void Arrangement::remove (double x, Segment *a, Eventheap &events, 
			  SegmentTree &sweepline)
{
  Segment *pred = sweepline.pred(a), *succ = sweepline.succ(a);
  sweepline.remove(a);
  backward_edge(a, a->head(), pred, succ);
  post_check(x, pred, succ, events);  
}

void Arrangement::swap (double x, Segment *a, Segment *b,
			Eventheap &events, SegmentTree &sweepline)
{
  if (b != sweepline.succ(a))
    return;
  double y = fabs(a->dy(x)) < fabs(b->dy(x)) ? a->y(x) : b->y(x);
  Point p(x, y);
  Segment *pred = sweepline.pred(a), *succ = sweepline.succ(b);
  sweepline.swap(a, b);
  backward_edge(a, p, pred, b);
  backward_edge(b, p, a, succ);
  forward_edge(b, p, pred, a);
  forward_edge(a, p, b, succ);
  post_check(x, pred, b, events);
  post_check(x, a, succ, events);
  if (a->circle || b->circle)
    post_check(x, b, a, events);
}

void Arrangement::check (double x, Segment *a, Segment *b, 
			 Eventheap &events, SegmentTree &sweepline)
{
  double nx;
  if (a->head().x <= x || b->head().x <= x || 
      b != sweepline.succ(a))
    return;
  if (x >= b->lb() && x >= a->lb() && b->below(a, x)) // changed from >
    events.push(Swap, x, a, b);
  else if (a->next(b, x, nx)) {
    events.push(Swap, nx, a, b);
  }
}

void Arrangement::post_check (double x, Segment *a, Segment *b, 
			      Eventheap &events)
{
  if (a && b)
    events.push(Check, x, a, b);
}

void Arrangement::forward_edge (Segment *s, const Point &p, Segment *pred,
				Segment *succ)
{
  Edge *e = new Edge(s, p, s->in);
  edges.push_back(e);
  s->edge = e;
  if (pred) {
    pred->edge->above.push_back(e);
    e->below.push_back(pred->edge);
  }
  if (succ) {
    succ->edge->below.push_back(e);
    e->above.push_back(succ->edge);
  }
}

void Arrangement::backward_edge (Segment *s, const Point &p, Segment *pred,
				 Segment *succ)
{
  Edge *e = new Edge(s, p, !s->in, s->edge);
  edges.push_back(e);
  s->edge->twin = e;
  if (pred) {
    pred->edge->above.push_back(s->edge);
    s->edge->below.push_back(pred->edge);
  }
  if (succ) {
    succ->edge->below.push_back(s->edge);
    s->edge->above.push_back(succ->edge);
  }
}

void Arrangement::set_heights ()
{
  int h = 1;
  for (unsigned int i = 0; i < edges.size(); ++i) {
    Edge *e = edges[i];
    if (e->inc() && e->above.empty())
      h = e->set_heights(h);
  }
}

void Arrangement::embed ()
{
  Edges in, out, vertical;
  EdgeSet sweep;
  int n = edges.size();
  for (int i = 0; i < n; ++i) {
    Edge *e = edges[i];
    int er = e->rank();
    if (er == 0)
      in.push_back(e);
    else if (er == 1)
      vertical.push_back(e);
    else if (er == 2)
      out.push_back(e);
    if (i == n - 1 || e->tail.x < edges[i+1]->tail.x) {
      embed_aux(in, out, vertical, sweep);
      in.clear();
      out.clear();
      vertical.clear();
    }
  }
}

/*
void Arrangement::embed_aux (Edges &in, Edges &out, Edges &vertical,
			     EdgeSet &sweep)
{
  for (unsigned int i = 0; i < in.size(); ++i)
    sweep.erase(in[i]);
  int nv;
  do {
    intersect_verticals(in, out, vertical, sweep);
    nv = vertical.size();
    adjust_endpoints(in, out, vertical, sweep);
  }
  while (nv < vertical.size());
  arrange_verticals(in, out, vertical);
  for (unsigned int i = 0; i < out.size(); ++i)
    sweep.insert(out[i]);
  add_embed_edges(in, out, vertical);
}
*/

void Arrangement::embed_aux (Edges &in, Edges &out, Edges &vertical,
			     EdgeSet &sweep)
{
  for (unsigned int i = 0; i < in.size(); ++i)
    sweep.erase(in[i]);
  intersect_verticals(in, out, vertical, sweep);
  adjust_endpoints(in, out, vertical, sweep);
  arrange_verticals(in, out, vertical);
  for (unsigned int i = 0; i < out.size(); ++i)
    sweep.insert(out[i]);
  add_embed_edges(in, out, vertical);
}

void Arrangement::intersect_verticals (Edges &in, Edges &out, Edges &vertical,
				       EdgeSet &sweep)
{
  for (unsigned int i = 0; i < vertical.size(); ++i) {
    Edge *v = vertical[i];
    EdgeIter is = edge_above(v->tail, in, out, sweep);
    EdgeIter ie = edge_above(v->twin->tail, in, out, sweep);
    if (is != sweep.end() && (ie == sweep.end() || **is < **ie))
      intersect_sweep(v->tail.x, is, ie, in, out, sweep);
    else if (ie != sweep.end() && (is == sweep.end() || **ie < **is))
      intersect_sweep(v->tail.x, ie, is, in, out, sweep);
  }
  intersect_verticals_degenerate(in, out, sweep);
}

void Arrangement::intersect_verticals_degenerate (Edges &in, Edges &out,
						  EdgeSet &sweep)
{
  Edges edges;
  for (unsigned int i = 0; i < in.size(); ++i)
    edges.push_back(in[i]);
  for (unsigned int i = 0; i < out.size(); ++i)
    edges.push_back(out[i]);
  sort(edges.begin(), edges.end(), edge_tail_order);
  for (unsigned int i = 0; i + 1 < edges.size(); ++i)
    if (edges[i]->tail.y == edges[i+1]->tail.y) {
      EdgeIter is = sweep.upper_bound(edges[i]);
      EdgeIter ie = sweep.upper_bound(edges[i+1]);
      if (is != sweep.end() && (ie == sweep.end() || **is < **ie))
	intersect_sweep(edges[i]->tail.x, is, ie, in, out, sweep);
      else if (ie != sweep.end() && (is == sweep.end() || **ie < **is))
	intersect_sweep(edges[i]->tail.x, ie, is, in, out, sweep);
    }
}

void Arrangement::intersect_sweep (double x, EdgeIter is, EdgeIter ie,
				   Edges &in, Edges &out, EdgeSet &sweep)
{
  while (is != ie) {
    Edge *e = *is;
    Point p(x, e->segment->y(x, false));
    Edge *f = split_edge(e, p);
    in.push_back(f->twin);
    out.push_back(e);
    EdgeIter os = is;
    ++is;
    sweep.erase(os);
  }
}

void Arrangement::adjust_endpoints (Edges &in, Edges &out, Edges &vertical,
				    EdgeSet &sweep)
{
  EdgeSet I, O;
  unsigned int i = 0, j = 0;
  while (i < in.size() || j < out.size()) {
    while (i < in.size()) {
      adjust_endpoint(in[i], I, vertical);
      pinch(in[i], I, O, in, out, sweep);
      ++i;
    }
    while (j < out.size()) {
      adjust_endpoint(out[j], O, vertical);
      pinch(out[j], I, O, in, out, sweep);
      ++j;
    }
  }
}

void Arrangement::adjust_endpoint (Edge *e, EdgeSet &E, Edges &vertical)
{
  EdgeIter ie = E.insert(e).first;
  adjust_endpoint1(ie, E, vertical);
  adjust_endpoint2(ie, E, vertical);
}

void Arrangement::adjust_endpoint1 (EdgeIter ie, EdgeSet &E, Edges &vertical)
{
  Point t = (*ie)->tail;
  EdgeIter ia = ie;
  ++ia;
  if (ia != E.end() && (*ia)->tail.y < t.y) {
    bool cflag = false;
    double de = fabs((*ie)->segment->dy(t.x));
    EdgeIter ib = ia;
    while (ib != E.end() && (*ib)->tail.y < t.y) {
      cflag = fabs((*ib)->segment->dy(t.x)) < de;
      if (cflag) break;
      ++ib;
    }
    Point p = cflag ? (*ib)->tail : t;
    EdgeIter i = cflag ? ie : ia;
    while (i != ib) {
      if ((*i)->tail.y != p.y)
	vertical.push_back(split_edge(*i, p));
      ++i;
    }
  }
}

void Arrangement::adjust_endpoint2 (EdgeIter ie, EdgeSet &E, Edges &vertical)
{
  Point t = (*ie)->tail;
  EdgeIter ia = ie;
  --ia;
  if (ie != E.begin() && (*ia)->tail.y > t.y) {
    bool cflag = false;
    double de = fabs((*ie)->segment->dy(t.x));
    EdgeIter ib = ia;
    while ((*ib)->tail.y > t.y) {
      cflag = fabs((*ib)->segment->dy(t.x)) < de;
      if (cflag || ib == E.begin())
	break;
      --ib;
    }
    Point p = cflag ? (*ib)->tail : t;
    if (cflag || (*ib)->tail.y <= t.y)
      ++ib;
    EdgeIter i = cflag ? ie : ia;
    while (true) {
      if ((*i)->tail.y != p.y)
	vertical.push_back(split_edge(*i, p));
      if (i == ib) break;
      --i;
    }
  }
}

void Arrangement::pinch (Edge *e, EdgeSet &I, EdgeSet &O, Edges &in,
			 Edges &out, EdgeSet &sweep)
{
  double x = e->tail.x;
  EdgeIter ep = sweep.upper_bound(e);
  if (ep != sweep.end()) {
    EdgeIter i = I.upper_bound(*ep), o = O.upper_bound(*ep);
    if (i != I.end() && (*i)->tail.y <= e->tail.y) {
      intersect_sweep(x, ep, sweep.upper_bound(*i), in, out, sweep);
      ep = sweep.upper_bound(e);
    }
    if (o != O.end() && (*o)->tail.y <= e->tail.y)
      intersect_sweep(x, ep, sweep.upper_bound(*o), in, out, sweep);
  }
  ep = sweep.upper_bound(e);
  if (ep == sweep.begin())
    return;
  EdgeIter em = ep;
  --em;
  EdgeIter i = I.upper_bound(*em), o = O.upper_bound(*em);
  if (i != I.begin()) {
    --i;
    if ((*i)->tail.y >= e->tail.y) {
      intersect_sweep(x, sweep.upper_bound(*i), ep, in, out, sweep);
      ep = sweep.upper_bound(e);
    }
  }
  if (o != O.begin()) {
    --o;
    if ((*o)->tail.y >= e->tail.y)
      intersect_sweep(x, sweep.upper_bound(*o), ep, in, out, sweep);
  }
}

void Arrangement::arrange_verticals (Edges &in, Edges &out, Edges &vertical)
{
  if (vertical.empty())
    return;
  unique_verticals(vertical);
  double x = vertical[0]->tail.x;
  set<double> y;
  for (unsigned int i = 0; i < in.size(); ++i)
    y.insert(in[i]->tail.y);
  for (unsigned int i = 0; i < out.size(); ++i)
    y.insert(out[i]->tail.y);
  int n = vertical.size();
  for (int i = 0; i < n; ++i) {
    y.insert(vertical[i]->tail.y);
    y.insert(vertical[i]->twin->tail.y);
  }
  for (int i = 0; i < n; ++i) {
    Edge *v = vertical[i];
    set<double>::iterator ys = y.upper_bound(v->tail.y);
    set<double>::iterator ye = y.find(v->twin->tail.y);
    while (*ys < *ye) {
      vertical.push_back(split_edge(v, Point(x, *ys)));
      ++ys;
    }
  }
  for (unsigned int i = 0; i < vertical.size(); ++i)
    vertical[i]->h = vertical[i]->twin->h = i;
}

Edge * Arrangement::split_edge (Edge *e, const Point &p)
{
  Edge *f = new Edge(e->segment, e->tail, e->in, 0, e->h);
  f->ns = e->ns;
  Edge *ft = new Edge(e->segment, p, !e->in, f, e->h);
  f->twin = ft;
  e->tail = p;
  edges.push_back(f);
  edges.push_back(ft);
  if (f->tail.x < p.x)
    sedges[e->ns] = f;
  return f->tail < f->twin->tail ? f : ft;
}

void Arrangement::add_embed_edges (Edges &in, Edges &out, Edges &vertical)
{
  Edges nedges(in);
  for (unsigned int i = 0; i < out.size(); ++i)
    nedges.push_back(out[i]);
  for (unsigned int i = 0; i < vertical.size(); ++i) {
    nedges.push_back(vertical[i]);
    nedges.push_back(vertical[i]->twin);
  }
  sort(nedges.begin(), nedges.end(), edge_order);

  for (int i = 0; i < nedges.size(); ++i)
    for (int j = i + 1; j < nedges.size(); ++j)
      if (nedges[i] == nedges[j])
	cerr << "same" << endl;

  int ns = sedges.size();
  for (unsigned int i = 0; i < nedges.size(); ++i) {
    Edge *e = nedges[i];
    e->ns = ns;
    ++ns;
    if (e->tail != e->twin->tail)
      sedges.push_back(e);
  }
}

void Arrangement::form_vertices ()
{
 int s = 0, e = 1, n = sedges.size();
  while (s < n) {
    if (e == n || sedges[s]->tail != sedges[e]->tail) {
      form_vertex(s, e);
      s = e;
    }
    ++e;
  }
}

void Arrangement::form_vertex (int s, int e)
{
  for (int i = s; i < e - 1; ++i)
    sedges[i]->next = sedges[i+1];
  sedges[e-1]->next = sedges[s];
}

void Arrangement::form_faces ()
{
  faces.push_back(new Face);
  int s = 0, e = 1, n = sedges.size();
  while (s < n) {
    if (e == n || sedges[s]->tail != sedges[e]->tail) {
      form_faces_vertex(s, e);
      s = e;
    }
    ++e;
  }
}

void Arrangement::form_faces_vertex (int s, int e)
{
  ++nv;
  bool traversed = false;
  for (int i = s; i < e; ++i)
    if (sedges[i]->traversed)
      traversed = true;
  if (!traversed) {
    ++nc;
    Face *enc = enclosing_face(sedges[s]->tail);
    enc->form_loop(sedges[s]);      
  }
  for (int i = s; i < e; ++i)
    if (!sedges[i]->traversed) {
      Face *f = new Face;
      faces.push_back(f);
      f->form_loop(sedges[i]);
    }
}

Face * Arrangement::enclosing_face (const Point &p, bool inflag)
{
  for (unsigned int i = 1; i < faces.size(); ++i)
    if (faces[i]->contains(p))
      return (inflag || faces[i]->in) ? faces[i] : 0;
  return faces[0];
}

void Arrangement::set_crossing_numbers (Face *f0, int cn0)
{
  int cnmax = 10 + faces.size();
  for (unsigned int i = 0; i < faces.size(); ++i)
    faces[i]->cn = cnmax;
  f0->cn = cn0;
  if (faces[0] != f0)
    faces[0]->cn = cnmax - 1;
  Faces st;
  st.push_back(f0);
  while (!st.empty()) {
    Face *f = *(st.end() - 1);
    st.pop_back();
    for (unsigned int j = 0; j < f->loops.size(); ++j) {
      Edge *e0 = f->loops[j], *e = e0->twin->next;
      do {
	Face *g = e->twin->face;
	if (g->cn == cnmax && !e->tiny()) {
	  g->cn = e->in ? f->cn + 1 : f->cn - 1;
	  st.push_back(g);
	}
	e = e->twin->next;
      }
      while (e != e0);      
    }
  }
}

bool Arrangement::planar ()
{
  return nv - sedges.size()/2 + faces.size() - 1 - nc == 0;
}

bool edge_order (Edge *a, Edge *b)
{
  if (a->tail < b->tail)
    return true;
  if (b->tail < a->tail)
    return false;
  int ar = a->rank(), br = b->rank();
  if (ar < br)
    return true;
  else if (ar > br)
    return false;
  return ar < 2  ? *a < *b : *b < *a;
}

bool edge_tail_order (Edge *a, Edge *b)
{
  return a->tail.y < b->tail.y;
}

EdgeIter edge_above (const Point &p, Edges &in, Edges &out, EdgeSet &sweep)
{
  for (unsigned int i = 0; i < in.size(); ++i)
    if (in[i]->tail == p)
      return sweep.upper_bound(in[i]);
  for (unsigned int i = 0; i < out.size(); ++i)
    if (out[i]->tail == p)
      return sweep.upper_bound(out[i]);
  Edge ep(0, p, true);
  return sweep.upper_bound(&ep);
}

void unique_verticals (Edges &vertical)
{
  Edges uvertical;
  for (unsigned int i = 0; i < vertical.size(); ++i) {
    bool flag = true;
    for (unsigned int j = i + 1; flag && j < vertical.size(); ++j)
      if (vertical[i]->tail == vertical[j]->tail &&
	  vertical[i]->twin->tail == vertical[j]->twin->tail &&
	  vertical[i]->in == !vertical[j]->in) {
	vertical[j] = *(vertical.end() - 1);
	vertical.pop_back();
	flag = false;
      }
    if (flag)
      uvertical.push_back(vertical[i]);
  }
  vertical = uvertical;
}

Arrangement * arrange (Segments &segments)
{
  Arrangement *res = new Arrangement;
  res->arrange(segments);
  if (!res->planar())
    cerr << "nonplanar" << endl;
  return res;
}

// debug

void pe (Edge *e)
{
  Points points;
  e->discretize(0.001, points);
  points[0] = e->tail;
  points[points.size() - 1] = e->twin->tail;
  cerr << "(";
  for (unsigned int i = 0; i < points.size(); ++i)
    cerr << "(" << points[i].x << " " << points[i].y << ")";
  cerr << ")" << endl;
}

void pl (Edge *e)
{
  if (!e) return;
  Edge *f = e;
  cerr << "(";
  do {
    pe(f);
    f = f->twin->next;
  }
  while (f != e);
  cerr << ")" << endl;
}
  
void pf (Face *f)
{
  cerr << setprecision(18);
  //cerr << "Face has " << f->loops.size() << " loops." << endl;
  cerr << "(";
  for (unsigned int i = 0; i < f->loops.size(); ++i) {
    //cerr << "Loop " << i << endl;
    pl(f->loops[i]);
  }
  cerr << ")";
}

void plp (Edge *e)
{
  if (!e) return;
  Edge *f = e;
  cerr << "(" << setprecision(18);
  do {
    cerr << "(" << f->tail.x << " " << f->tail.y << ")";
    f = f->twin->next;
  }
  while (f != e);
  cerr << ")" << endl;
}

void pp (const Points &points)
{
  cerr << "(";
  for (unsigned int i = 0; i < points.size(); ++i)
    cerr << "(" << points[i].x << " " << points[i].y << ") ";
  cerr << ")" << endl;
}

void ps (Segment *s)
{
  if (s->circle)
    cerr << "circle: t = (" << s->tail().x << " " << s->tail().y << ") h = ("
	 << s->head().x << " " << s->head().y << ") in = " << s->in 
	 << " o = (" << s->o.x << " " << s->o.y << ") r = "
	 << s->r << endl;
  else
    cerr << "line: t = (" << s->tail().x << " " << s->tail().y << ") h = ("
	 << s->head().x << " " << s->head().y << ") in = " << s->in << endl;
}

void psa (Segment *s)
{
  cerr << "(" << setprecision(18);
  for (unsigned int i = 0; i < s->np; ++i)
    cerr << "(" << s->points[i].x << " " << s->points[i].y << ") ";
  cerr << ")" << endl;
}

void write_segments (const Segments &segments, ostream &ostr)
{
  ostr << setprecision(18) << "(";
  for (unsigned int i = 0; i < segments.size(); ++i) {
    ostr << "(";
    Point *points = segments[i]->points;
    for (int j = 0; j < segments[i]->np; ++j)
      ostr << "(" << points[j].x << " " << points[j].y << ") ";
    ostr << ")" << endl;
  }
  ostr << ")" << endl;
}

void ws (const Segments &segments)
{
  write_segments(segments, cerr);
}

void write_segments2 (const Segments &segments, ostream &ostr, double tol)
{
  ostr << setprecision(18) << "(";
  for (unsigned int i = 0; i < segments.size(); ++i) {
    Points points;
    Segment *c = segments[i];
    c->discretize(c->tail().x, c->head().x, tol, points);
    points[0] = c->tail();
    points[points.size() - 1] = c->head();
    ostr << "(";
    for (unsigned int j = 0; j < points.size(); ++j)
      ostr << "(" << points[j].x << " " << points[j].y << ") ";
    ostr << ")" << endl;
  }
  ostr << ")" << endl;
}

void write_segments3 (const Segments &segments, ostream &ostr)
{
  ostr << setprecision(18) << segments.size() << endl;
  for (unsigned int i = 0; i < segments.size(); ++i) {
    ostr << segments[i]->np << endl;
    Point *points = segments[i]->points;
    for (int j = 0; j < segments[i]->np; ++j)
      ostr << points[j].x << " " << points[j].y << " ";
    ostr << endl;
  }
  ostr << endl;
}

void ws3 (const Segments &segments)
{
  write_segments3(segments, cerr);
}

void read_segments (Segments &segments, istream &istr)
{
  int n;
  istr >> n;
  for (int i = 0; i < n; ++i) {
    Points points;
    read_points(points, istr);
    segments.push_back(new Segment(points, true));
  }
}

void minimal_inconsistent_set (const Segments &segs, Segments &bad)
{
  Segments osegs = segs, nsegs;
  while (true) {
    minimal_inconsistent_set_aux(osegs, nsegs);
    if (osegs.size() == nsegs.size())
      break;
    osegs = nsegs;
    nsegs.clear();
  }
  bad = nsegs;
}

void minimal_inconsistent_set_aux (const Segments &segs, Segments &bad)
{
  bad = segs;
  for (int i = 0; i < segs.size(); ++i) {
    Segments nsegs = bad;
    Segments::iterator iter = remove(nsegs.begin(), nsegs.end(), segs[i]);
    nsegs.erase(iter, nsegs.end());
    Arrangement *arr = new Arrangement;
    arr->arrange(nsegs);
    if (!arr->planar())
      bad = nsegs;
    delete arr;
  }
}

/*
main ()
{
  Segments segs;
  read_segments(segs, cin);
  Arrangement *arr = arrange(segs);
  delete_segments(segs);
  delete arr;
}
*/
