#ifndef ARRANGE
#define ARRANGE

#include "point.h"
#include <float.h>
#include <set>

#ifdef WIN32
#define M_PI 3.14159265358979323846
#endif

class Rootlist {
 public:
  Rootlist (bool ibelow = true) : below(ibelow), n(0) {};

  bool below;
  int n;
  double x[2];
};

class Segment;

class SegmentNode {
 public:
  SegmentNode (Segment *is = 0) : s(is) {}
  Segment * pred (Segment *s) const;
  Segment * succ (Segment *s) const;

  Segment *s;
};

class SegmentOrder {
 public:
  bool operator() (SegmentNode *a, SegmentNode *b) const;
};

typedef set<SegmentNode *, SegmentOrder> SegmentSet;

class Edge;

class Splat;

class Segment {
 public:
  Segment () : np(0), points(0), edge(0), splat(0) {}
  Segment (const Point &t, const Point &h, bool iin, Splat *isplat = 0);
  Segment (const Point &t, const Point &h, const Point &io,
	   double ir, bool iupper, bool iin);
  Segment (const Points &ipoints, bool iin);
  ~Segment () { delete [] points; }
  bool vertical () const { return points[0].x == points[np-1].x; }
  Point tail () const { return points[0]; }
  Point head () const { return points[np-1]; }
  void shift () { ++ip; }
  double lb () const { return points[ip].x; }
  double ub () const { return points[ip+1].x; }
  double ymax () const;
  double ymin () const;
  double y (double x, bool ipflag = true) const;
  double dy (double x) const;
  Rootlist rootlist (const Segment &s) const;
  Rootlist rootlist_g (const Segment &s) const;
  Rootlist rootlist_ll (const Segment &s) const;
  int intersection (const Segment &s, double *x) const;
  int intersection_ll (const Segment &s, double *x) const;
  int intersection_cl (const Segment &s, double *x) const;
  int intersection_cc (const Segment &s, double *x) const;
  bool next (Segment *s, double x, double &nx) const;
  bool below (Segment *s, double x) const;
  void bounding_box (double xs, double xe, double *bbox) const;
  void discretize (double xs, double xe, double tol, Points &res) const;
  void discretize_circle (double xs, double xe, double tol, Points &res) const;
  void discretize_pline (double xs, double xe, Points &res) const;

  bool circle, upper, in;
  int np, ip;
  Point *points, o;
  double r;
  Edge *edge;
  SegmentSet::iterator node;
  Splat *splat;
};

typedef vector<Segment *> Segments;

void delete_segments (Segments &segments);

class SegmentTree {
 public:
  SegmentTree () {}

  void insert (Segment *s);
  void remove (Segment *s);
  void swap (Segment *s, Segment *t);
  Segment * succ (Segment *s) const;
  Segment * pred (Segment *s) const;

  SegmentSet data;
};

double line_y (double x, const Point &p, const Point &q);

bool line_intersection (double a1, double b1, double c1, double a2,
			double b2, double c2, double *res);

bool circle_line_intersections (double ox, double oy, double r,
				double a, double b, double c, double *res);

bool circle_intersections (double x1, double y1, double r1, 
			   double x2, double y2, double r2, 
			   double *res, bool safe = false);

enum EventType {Insert, Shift, Remove, Swap, Check};

class Event {
 public:
   Event () : a(0), b(0) {}
  Event (EventType itype, double ix, Segment *ia, Segment *ib = 0)
    : type(itype), x(ix), a(ia), b(ib) {}
  bool operator< (const Event &b) const;
  
  EventType type;
  double x;
  Segment *a, *b;
};

class Eventheap {
 public:
  Eventheap () : n(0), nmax(0), events(0) {}
  Eventheap (const Segments &segments);
  ~Eventheap () { delete [] events; }
  bool empty () const { return n == 0; }
  Event top () const { return events[0]; }
  inline Event * pop ();
  inline void push (EventType type, double x, Segment *a, Segment *b = 0);

  int n, nmax;
  Event *events;
};

class Face;

class Edge {
 public:
  Edge (Segment *isegment, const Point &itail, bool iin, Edge *itwin = 0,
	int ih = 0)
    : segment(isegment), tail(itail), in(iin), traversed(false), twin(itwin),
    next(0), face(0), h(ih) {}
  bool inc () const { return tail.x < twin->tail.x; }
  bool tiny () const {return tail.dist2(twin->tail) < 1e-14; }
  bool operator< (const Edge &b) const { return h < b.h; }
  int set_heights (int h0);
  int rank () const;
  bool contains (const Point &p) const;
  bool above_point (const Point &p) const;
  void discretize (double tol, Points &res) const;
  void restrict (Segments &res) const;
  bool null_loop () const;
  int loop_length () const;
  void bounding_box (double *bbox) const;

  Segment *segment;
  Point tail;
  bool in, traversed;
  Edge *twin, *next;
  vector<Edge *> below, above;
  Face *face;
  int h, ns;
};

typedef vector<Edge *> Edges;

class EdgeOrder {
 public:
  bool operator() (Edge *a, Edge *b) const
    { return a->segment ? *a < *b : a->tail.y < b->segment->y(a->tail.x, false); }
};

typedef set<Edge *, EdgeOrder> EdgeSet;

typedef EdgeSet::iterator EdgeIter;

class Face {
 public:
  Face () : in(true), visited(false) {}
  void form_loop (Edge *e);
  void traverse (Profile &boundary) const;
  void discretize (double tol, Profile &boundary, double dmin = 0.0) const;
  bool contains (const Point &p) const;

  Edges loops;
  bool in, visited;
  int cn;
  double bbox[4];
};

typedef vector<Face *> Faces;

bool bbox_contains (const double *bbox, const Point &p);

class Arrangement {
 public:
  Arrangement () : nv(0), nc(0) {}
  ~Arrangement ();
  void arrange (Segments &segments);
  void split_near_vertical (Segments &segments) const;
  void sweep (const Segments &segments);
  void insert (double x, Segment *a, Eventheap &events, SegmentTree &sweepline);
  void shift (double x, Segment *a, Eventheap &events, SegmentTree &sweepline);
  void remove (double x, Segment *a, Eventheap &events, SegmentTree &sweepline);
  void swap (double x, Segment *a, Segment *b, Eventheap &events, 
	     SegmentTree &sweepline);
  void check (double x, Segment *a, Segment *b, Eventheap &events, 
	      SegmentTree &sweepline);
  void post_check (double x, Segment *a, Segment *b, Eventheap &events);
  void forward_edge (Segment *s, const Point &p, Segment *pred = 0,
		     Segment *succ = 0);
  void backward_edge (Segment *s, const Point &p, Segment *pred = 0,
		      Segment *succ = 0);
  void set_heights ();
  void embed ();
  void embed_aux (Edges &in, Edges &out, Edges &vertical, EdgeSet &sweep);
  void intersect_verticals (Edges &in, Edges &out, Edges &vertical,
			    EdgeSet &sweep);
  void intersect_verticals_degenerate (Edges &in, Edges &out, EdgeSet &sweep);
  void intersect_sweep (double x, EdgeIter is, EdgeIter ie, Edges &in, 
			Edges &out, EdgeSet &sweep);
  void adjust_endpoints (Edges &in, Edges &out, Edges &vertical,
			 EdgeSet &sweep);
  void adjust_endpoint (Edge *e, EdgeSet &E, Edges &vertical);
  void adjust_endpoint1 (EdgeIter ie, EdgeSet &E, Edges &vertical);
  void adjust_endpoint2 (EdgeIter ie, EdgeSet &E, Edges &vertical);
  void pinch (Edge *e, EdgeSet &I, EdgeSet &O, Edges &in, Edges &out,
	      EdgeSet &sweep);
  void arrange_verticals (Edges &in, Edges &out, Edges &verticals);
  Edge * split_edge (Edge *e, const Point &p);
  void add_embed_edges (Edges &in, Edges &out, Edges &vertical);
  void form_vertices ();
  void form_vertex (int s, int e);
  void form_faces ();
  void form_faces_vertex (int s, int e);
  Face * enclosing_face (const Point &p, bool inflag = true);
  void set_crossing_numbers (Face *f0, int cn0);
  bool planar ();

  Edges edges, sedges;
  Faces faces;
  int nv, nc;
};

bool edge_order (Edge *a, Edge *b);

bool edge_tail_order (Edge *a, Edge *b);

EdgeIter edge_above (const Point &p, Edges &in, Edges &out, EdgeSet &sweep);

void unique_verticals (Edges &vertical);

Arrangement * arrange (Segments &segments);

// debug

void pe (Edge *e);

void pl (Edge *e);

void pf (Face *f);

void plp (Edge *e);

void pp (const Points &points);

void ps (Segment *s);

void psa (Segment *s);

void write_segments (const Segments &segments, ostream &ostr);

void ws (const Segments &segments);

void write_segments2 (const Segments &segments, ostream &ostr, double tol);

void write_segments3 (const Segments &segments, ostream &ostr);

void ws3 (const Segments &segments);

void read_segments (Segments &segments, istream &istr);

void minimal_inconsistent_set (const Segments &segs, Segments &bad);

void minimal_inconsistent_set_aux (const Segments &segs, Segments &bad);

#endif
