#include "cs.h"

double ContactFunction::y (double x, bool &flag) const
{
  switch (type) {
  case rara: case rarl: case rlra:
    return y_rr(x, flag);
  case rata:
    return y_rata(x, flag);
  case ratl:
    return p[1]*cos(x + p[0]) + p[2];
  case rlta:
    return y_rlta(x, flag);
  case tata:
    return y_tata(x, flag);
  case linear:
    return p[1] == 0.0 ? - p[2]/p[0] : - (p[0]*x + p[2])/p[1];
  }
  return 0.0;
}

double ContactFunction::y_rr (double x, bool &flag) const
{
  double s = sin(x), c = cos(x), th[2], u, v, w;
  u = p[0]*s + p[1]*c + p[2];
  v = p[0]*c - p[1]*s + p[3];
  w = p[4]*s + p[5]*c + p[6];
  flag = sin_cos_roots(u, v, w, 0.0, th);
  return p[7] == 0.0 ? th[0] : th[1];
}

double ContactFunction::y_rata (double x, bool &flag) const
{
  double temp1, temp2, temp3;
  temp1 = p[5] - p[1]*sin(x + p[2]);
  temp2 = p[4] - temp1*temp1;
  if (temp2 < 0.0) {
    flag = false;
    temp3 = 0.0;
  }
  else
    temp3 = p[3]*sqrt(temp2);
  return p[0] + p[1]*cos(x + p[2]) + temp3;
}

double ContactFunction::y_rlta (double x, bool &flag) const
{
  double temp = sin(x + p[1]);
  if (temp == 0.0) {
    flag = false;
    temp = 1e-10;
  }
  return p[3] + (p[0]*cos(x + p[1]) - p[2])/temp;
}

double ContactFunction::y_tata (double x, bool &flag) const
{
  double x1 = p[3] - p[2], x2 = p[3] + p[2];
  if (x == x1 || x == x2)
    return p[0];
  if (x < x1 && x < x2 || x > x1 && x > x2) {
    flag = false;
    return p[0];
  }
  double d = p[2]*p[2] - (p[3] - x)*(p[3] - x);
  if (d < 0.0) {
    flag = false;
    return p[0];
  }
  d = sqrt(d);
  return p[0] + p[1]*d;
}

void ContactFunction::rara_domain (Domain &res) const
{
  if (vertical()) return;
  int n = 0;
  double x[4];
  if (sin_cos_roots(p[4], p[5], p[8], 0.0, x, 1e-10))
    n = 2;
  if (sin_cos_roots(p[4], p[5], p[9], 0.0, x + n, 1e-10))
    n += 2;
  domain_aux(n, x, res);
}

void ContactFunction::rarl_domain (Domain &res) const
{
  double x[2], u = p[0]*p[2] - p[1]*p[3], v = p[1]*p[2] + p[0]*p[3],
    w = 0.5*(p[0]*p[0] + p[1]*p[1] + p[2]*p[2] + p[3]*p[3] - p[6]*p[6]);
  int n = sin_cos_roots(u, v, w, 0.0, x) ? 2 : 0;
  domain_aux(n, x, res);
}

void ContactFunction::rlra_domain (Domain &res) const
{
  if (vertical()) return;
  int n = 0;
  double x[4], u = p[4], v = p[5], d = sqrt(p[0]*p[0] + p[1]*p[1]),
    w1 = p[6] + d, w2 = p[6] - d;
  if (sin_cos_roots(u, v, w1, 0.0, x))
    n = 2;
  if (sin_cos_roots(u, v, w2, 0.0, x + n))
    n += 2;
  domain_aux(n, x, res);
}

void ContactFunction::rata_domain (Domain &res) const
{
  int n = 0;
  double x[4];
  if (p[1] != 0.0) {
    double k = sqrt(p[4]), k1 = (p[5] + k)/p[1], k2 = (p[5] - k)/p[1];
    if (sin_roots(k1, p[2], x))
      n = 2;
    if (sin_roots(k2, p[2], x + n))
      n += 2;
  }
  domain_aux(n, x, res);
}

void ContactFunction::domain_aux (int n, double *x, Domain &res) const
{
  sort(x, x + n);
  for (int i = 0; i < n; ++i) {
    int j = i == n - 1 ? 0 : i + 1;
    if (x[i] == x[j])
      continue;
    double xm = j == 0 ? mod_pi(0.5*(x[i] + x[j] + 2.0*M_PI)) 
      : 0.5*(x[i] + x[j]);
    bool flag = true;
    y(xm, flag);
    if (flag) {
      res.push_back(x[i]);
      res.push_back(x[j]);
    }
  }
}

Point domain_point (const ContactFunction *f, double x)
{
  return Point(x, 0.5*(f[0].y(x) + f[1].y(x)));
}

bool domain_order (double x1, double x2, const Domain &domain)
{
  if (domain[0] < domain[1])
    return x1 < x2;
  double xm = 0.5*(domain[0] + domain[1]);
  if (x1 < xm && x2 > xm)
    return false;
  if (x1 > xm && x2 < xm)
    return true;
  return x1 < x2;
}

void ContactInterval::intersect (const ContactInterval &bint,
				 const ContactFunction *f, 
				 ContactIntervals &ints) const
{
  if (f->vertical())
    intersect_v(bint, f, ints);
  else
    intersect_g(bint, f, ints);
}

void ContactInterval::intersect_v (const ContactInterval &bint,
				   const ContactFunction *f, 
				   ContactIntervals &ints) const
{
  ContactIntervals aints, bints;
  simple_intervals_v(f, aints);
  bint.simple_intervals_v(f, bints);
  for (unsigned int i = 0; i < aints.size(); ++i) {
    const ContactInterval &ai = aints[i];
    for (unsigned int j = 0; j < bints.size(); ++j) {
      const ContactInterval &bj = bints[j];
      if (ai.tlower == bj.tlower) {
	const Point &lb = ai.t.y < bj.t.y ? bj.t : ai.t;
	const Point &ub = ai.h.y < bj.h.y ? ai.h : bj.h;
	if (lb.y < ub.y)
	  ints.push_back(ContactInterval(lb, ub, ai.tlower));
      }
    }
  }
}

void ContactInterval::simple_intervals_v (const ContactFunction *f,
					  ContactIntervals &ints) const
{
  if (t == h)
    return;
  if (t.y < h.y)
    ints.push_back(*this);
  else {
    const ContactFunction &g = tlower ? f[0] : f[1];
    Point p(g.y(0.0), M_PI), q(p.x, - M_PI);
    ints.push_back(ContactInterval(t, p, tlower));
    ints.push_back(ContactInterval(q, h, tlower));
  }
}

void ContactInterval::intersect_g (const ContactInterval &bint,
				   const ContactFunction *f,
				   ContactIntervals &ints) const
{
  ContactIntervals aints, bints;
  simple_intervals_g(f, aints);
  bint.simple_intervals_g(f, bints);
  for (unsigned int i = 0; i < aints.size(); ++i) {
    const ContactInterval &ai = aints[i];
    for (unsigned int j = 0; j < bints.size(); ++j) {
      const ContactInterval &bj = bints[j];
      if (ai.tlower == bj.tlower) {
	const Point &lb = ai.t < bj.t ? bj.t : ai.t;
	const Point &ub = ai.h < bj.h ? ai.h : bj.h;
	if (lb.x < ub.x)
	  ints.push_back(ContactInterval(lb, ub, ai.tlower));
      }
    }
  }
}

void ContactInterval::simple_intervals_g (const ContactFunction *f,
					  ContactIntervals &ints) const
{
  if (t == h || t.x == M_PI && h.x == - M_PI)
    return;
  Point nt = t, nh = h;
  if (nt.x == M_PI)
    nt.x = - nt.x;
  if (nh.x == - M_PI)
    nh.x = - nh.x;
  if (nt < nh)
    ints.push_back(ContactInterval(nt, nh, tlower));
  else {
    const ContactFunction &g = tlower ? f[0] : f[1];
    Point p(M_PI, g.y(M_PI)), q(- p.x, p.y);
    ints.push_back(ContactInterval(nt, p, tlower));
    ints.push_back(ContactInterval(q, nh, tlower));
  }
}

void ContactInterval::intersect (const ContactInterval &bint,
				 const ContactFunction *f, const Domain &domain,
				 ContactIntervals &ints) const
{
  ContactIntervals sints;
  if (t == h && tlower && !hlower)
    bint.simple_intervals(f, domain, sints);
  else if (bint.t == bint.h && bint.tlower && !bint.hlower)
    simple_intervals(f, domain, sints);
  else {
    intersect_ep(t, tlower, bint.t, bint.tlower, ints);
    intersect_ep(h, hlower, bint.h, bint.hlower, ints);
    intersect_ep(t, tlower, bint.h, bint.hlower, ints);
    intersect_ep(h, hlower, bint.t, bint.tlower, ints);
    ContactIntervals rints;
    if (!(tlower == bint.hlower && t.x == bint.h.x) &&
	bint.contains(t, tlower, domain))
      if (contains(bint.h, bint.hlower, domain))
	rints.push_back(ContactInterval(t, tlower, bint.h, bint.hlower));
      else
	rints.push_back(*this);
    if (t != bint.t && !(hlower == bint.tlower && h.x == bint.t.x) &&
	contains(bint.t, bint.tlower, domain))
      if (bint.contains(h, hlower, domain))
	rints.push_back(ContactInterval(bint.t, bint.tlower, h, hlower));
      else
	rints.push_back(bint);
    for (unsigned int i = 0; i < rints.size(); ++i)
      rints[i].simple_intervals(f, domain, sints);
  }
  for (unsigned int i = 0; i < sints.size(); ++i)
    sints[i].simple_intervals_g(f, ints);
}

void intersect_ep (const Point &a, bool alower, const Point &b, bool blower,
		   ContactIntervals &ints)
{
  if (alower == blower && a.x == b.x)
    if (a.y < b.y)
      ints.push_back(ContactInterval(a, b, alower));
    else if (a.y > b.y)
      ints.push_back(ContactInterval(b, a, alower));
}

bool ContactInterval::contains (const Point &p, bool plower,
				const Domain &domain) const
{
  if (p.x == t.x && plower == tlower)
    return t.y <= p.y;
  if (p.x == h.x && plower == hlower)
    return h.y <= p.y;
  if (tlower && hlower && domain_order(t.x, h.x, domain))
    return plower && in_arc(p.x, t.x, h.x);
  else if (!tlower && !hlower && domain_order(h.x, t.x, domain))
    return !plower && in_arc(p.x, h.x, t.x);
  else if (tlower && hlower)
    return (!plower || domain_order(t.x, p.x, domain) ||
	    domain_order(p.x, h.x, domain));
  else if (!tlower && !hlower)
    return (plower || domain_order(p.x, t.x, domain) ||
	    domain_order(h.x, p.x, domain));
  else if (tlower)
    return plower ? domain_order(t.x, p.x, domain) :
      domain_order(h.x, p.x, domain);
  else
    return plower ? domain_order(p.x, h.x, domain) :
      domain_order(p.x, t.x, domain);
}

void ContactInterval::simple_intervals (const ContactFunction *f,
					const Domain &domain,
					ContactIntervals &ints) const
{
  if (tlower && hlower && domain_order(t.x, h.x, domain))
    ints.push_back(ContactInterval(t, h, true));
  else if (!tlower && !hlower && domain_order(h.x, t.x, domain))
    ints.push_back(ContactInterval(h, t, false));
  else {
    Point lb = domain_point(f, domain[0]), ub = domain_point(f, domain[1]);
    if (tlower && hlower) {
      ints.push_back(simple_interval(lb, h, true, domain));
      ints.push_back(simple_interval(t, ub, true, domain));
      ints.push_back(ContactInterval(lb, ub, false));
    }
    else if (!tlower && !hlower) {
      ints.push_back(simple_interval(lb, t, false, domain));
      ints.push_back(simple_interval(h, ub, false, domain));
      ints.push_back(ContactInterval(lb, ub, true));
    }
    else if (tlower) {
      ints.push_back(simple_interval(t, ub, true, domain));
      ints.push_back(simple_interval(h, ub, false, domain));
    }
    else {
      ints.push_back(simple_interval(lb, t, false, domain));
      ints.push_back(simple_interval(lb, h, true, domain));
    }
  }
}

ContactInterval simple_interval (const Point &t, const Point &h, 
				 bool lower, const Domain &domain)
{
  return (t.x == h.x && t.y < h.y || domain_order(t.x, h.x, domain)) ?
    ContactInterval(t, h, lower) : ContactInterval(h, t, lower);
}

void Feature::set_normals ()
{
  if (circle)
    if (in) {
      nt = Angle(t.x - o.x, t.y - o.y);
      nh = Angle(h.x - o.x, h.y - o.y);
    }
    else {
      nt = Angle(o.x - t.x, o.y - t.y);
      nh = Angle(o.x - h.x, o.y - h.y);
    }
  else {
    if (in)
      nt = Angle(h.y - t.y, t.x - h.x);
    else
      nt = Angle(t.y - h.y, h.x - t.x);
    nh = nt;
  }
}

Point Feature::point (const Angle &th) const
{
  if (th == nt)
    return t;
  if (th == nh)
    return h;
  return o + (in ? r : - r)*th;
}

void Feature::split_vertical (Angles &ints) const
{
  Angle zero(1.0, 0.0), pi(-1.0, 0.0);
  if (full_circle()) {
    ints.push_back(zero);
    ints.push_back(pi);
    ints.push_back(pi);
    ints.push_back(zero);
    return;
  }
  Angles tmp;
  tmp.push_back(nt);
  if (zero.in_arc_strict(nt, nh)) {
    tmp.push_back(zero);
    tmp.push_back(zero);
  }
  tmp.push_back(nh);
  for (unsigned int i = 0; i < tmp.size(); i +=2) {
    ints.push_back(tmp[i]);
    if (pi.in_arc_strict(tmp[i], tmp[i+1])) {
      ints.push_back(pi);
      ints.push_back(pi);
    }
    ints.push_back(tmp[i+1]);
  }
}

bool Feature::merge (Feature *f)
{
  if (!(in == f->in && circle == f->circle &&
	(!circle || o == f->o && r == f->r)))
    return false;
  if (h == f->t && nh == f->nt) {
    h = f->h;
    nh = f->nh;
    return true;
  }
  if (t == f->h && nt == f->nh) {
    t = f->t;
    nt = f->nt;
    return true;
  }
  return false;
}

void Feature::trim (Feature *f)
{
  Point &p = in ? h : t;
  Point &q = f->in ? f->t : f->h;
  if (p == q) return;
  Point r = closest_point(f, 0.5*(p + q));
  p = q = r;
  if (circle) set_normals();
  if (f->circle) f->set_normals();
}

Point Feature::closest_point (Feature *f, const Point &p) const
{
  double xy[4], a, b, c, u, v, w;
  int n;
  if (!circle)
    line_equation(a, b, c);
  if (!f->circle)
    f->line_equation(u, v, w);
  if (circle && f->circle)
    n = circle_intersections(o.x, o.y, r, f->o.x, f->o.y, f->r, xy) ? 2 : 0;
  else if (circle)    
    n = circle_line_intersections(o.x, o.y, r, u, v, w, xy) ? 2 : 0;
  else if (f->circle)
    n = circle_line_intersections(f->o.x, f->o.y, f->r, a, b, c, xy) ? 2 : 0;
  else
    n = line_intersection(a, b, c, u, v, w, xy);
  if (n == 0) 
    return p;
  Point q(xy[0], xy[1]);
  double d = p.dist2(q);
  for (int i = 1; i < n; ++i) {
    Point qi(xy[2*i], xy[2*i+1]);
    double di = p.dist2(qi);
    if (di < d) {
      d = di;
      q = qi;
    }
  }
  return q;
}

void Feature::line_equation (double &a, double &b, double &c, bool flag) const
{
  double dx, dy, den;
  bool inflag = flag ? in : !in;
  dx = inflag ? t.x - h.x : h.x - t.x;
  dy = inflag ? h.y - t.y : t.y - h.y;
  den = 1.0/sqrt(dx*dx + dy*dy);
  a = dy*den;
  b = dx*den;
  c = -a*t.x - b*t.y;
}

Feature Feature::move (double th, double x, double y) const
{
  Angle a(th);
  Point d(x, y);
  return move(a, d);
}

Feature Feature::move (const Angle &a, const Point &d) const

{
  if (circle)
    return Feature(a*t + d, a*h + d, a*o + d, r, in, a+nt, a+nh);
  else
    return Feature(a*t + d, a*h + d, in, a+nt, a+nh);
}

bool Feature::compatible (Feature *b) const
{
  if (!circle && t == h || !b->circle && b->t == b->h)
    return false;
  if (circle && b->circle && r == 0.0 && b->r == 0.0)
    return false;
  return true;
}

double Feature::aa_r (const Feature &arc) const
{
  double ra = in ? r : - r;
  double rb = arc.in ? arc.r : - arc.r;
  return ra + rb;
}

int Feature::rara_function (Feature *arc, const Point &q, ContactFunction *f) const
{
  const Point &ob = arc->o;
  if (o.x == 0.0 && o.y == 0.0 && ob.x == 0.0 && ob.y == 0.0)
    return 0;
  bool flag = in == arc->in;
  double sr = aa_r(*arc);
  double k1 = -0.5*(ob%ob + q%q + o%o - sr*sr);
  if (ob.x == 0.0 && ob.y == 0.0) {
    double th[2];
    if (!sin_cos_roots(o*q, o%q, k1, 0.0, th))
      return 0;
    for (int i = 0; i < 2; ++i) {
      f[i].type = linear;
      f[i].p[0] = 1.0;
      f[i].p[1] = 0.0;
      f[i].p[2] = - th[i];
      f[i].in = i == 0 ? !flag : flag;
    }
    return 2;
  }
  double k2 = 0.5*(sr*sr + ob%ob - q%q - o%o) + sr*ob.norm();
  double k3 = 0.5*(sr*sr + ob%ob - q%q - o%o) - sr*ob.norm();
  for (int i = 0; i < 2; ++i) {
    f[i].type = rara;
    f[i].p[0] = o%ob;
    f[i].p[1] = - o*ob;
    f[i].p[2] = - ob*q;
    f[i].p[3] = - ob%q;
    f[i].p[4] = o*q;
    f[i].p[5] = o%q;
    f[i].p[6] = k1;
    f[i].p[7] = i;
    f[i].p[8] = k2;
    f[i].p[9] = k3;
    f[i].in = i == 0 ? flag : !flag;
  }
  return 2;
}

int Feature::rarl_function (Feature *line, const Point &q, ContactFunction *f) 
  const
{
  double c;
  Point n;
  line->line_equation(n.x, n.y, c, in);
  for (int i = 0; i < 2; ++i) {
    f[i].type = rarl;
    f[i].p[0] = n%o;
    f[i].p[1] = n*o;
    f[i].p[2] = - n*q;
    f[i].p[3] = - n%q;
    f[i].p[4] = 0.0;
    f[i].p[5] = 0.0;
    f[i].p[6] = - r - n%line->t;
    f[i].p[7] = i;
    f[i].in = i == 1;
  }
  return 2;
}

int Feature::rlra_function (Feature *arc, const Point &q, ContactFunction *f)
  const
{
  double c, th[2];
  Point n, &o = arc->o;
  line_equation(n.x, n.y, c, arc->in);
  if (o.x == 0.0 && o.y == 0.0) {
    if (!sin_cos_roots(n*q, n%q, - arc->r - n%t, 0.0, th))
      return 0;
    for (int i = 0; i < 2; ++i) {
      f[i].type = linear;
      f[i].p[0] = 1.0;
      f[i].p[1] = 0.0;
      f[i].p[2] = - th[i];
      f[i].in = i == 0;
    }
    return 2;
  }
  for (int i = 0; i < 2; ++i) {
    f[i].type = rlra;
    f[i].p[0] = n%o;
    f[i].p[1] = - n*o;
    f[i].p[2] = 0.0;
    f[i].p[3] = 0.0;
    f[i].p[4] = n*q;
    f[i].p[5] = n%q;
    f[i].p[6] = - arc->r - n%t;
    f[i].p[7] = i == 0 ? 1.0 : 0.0;
    f[i].in = i == 0;
  }
  return 2;
}

int Feature::rata_function (Feature *arc, ContactFunction *f) const
{
  double alpha, d, s;
  o.polar(d, alpha);
  s = fabs(aa_r(*arc));
  if (d == 0.0 && fabs(arc->o.y) >= s)
    return 0;
  bool flag = in == arc->in;
  for (int i = 0; i < 2; ++i) {
    f[i].type = rata;
    f[i].p[0] = - arc->o.x;
    f[i].p[1] = d;
    f[i].p[2] = alpha;
    f[i].p[3] = i == 0 ? -1.0 : 1.0;
    f[i].p[4] = s*s;
    f[i].p[5] = arc->o.y;
    f[i].in = i == 0 ? flag : !flag;
  }
  return 2;
}

int Feature::ratl_function (Feature *line, ContactFunction *f) const
{
  double a, b, c, d, alpha, beta;
  line->line_equation(a, b, c, in);
  o.polar(d, alpha);
  beta = atan2(- b, - a);
  if (fabs(a) < EPSILON) {
    double th[2];
    if (d == 0.0 || !cos_roots((c-r)/d, alpha - beta, th))
      return 0;
    for (int i = 0; i < 2; ++i) {
      f[i].type = linear;
      f[i].p[0] = 1.0;
      f[i].p[1] = 0.0;
      f[i].p[2] = - th[i];
      f[i].in = i == 0;
    }
    return 2;
  }
  f->type = ratl;
  f->p[0] = alpha - beta;
  f->p[1] = - d/a;
  f->p[2] = (c - r)/a;
  f->in = a > 0.0;
  return 1;
}

int Feature::rlta_function (Feature *arc, ContactFunction *f) const
{
  double a, b, c;
  line_equation(a, b, c, arc->in);
  f->type = rlta;
  f->p[0] = arc->o.y;
  f->p[1] = atan2(- a, b);
  f->p[2] = - c + arc->r;
  f->p[3] = - arc->o.x;
  return 1;
}

ContactFunction Feature::tata_function (Feature *arc, bool upper) const
{
  ContactFunction f;
  double sr = aa_r(*arc);
  bool in = sr > 0.0;
  f.type = tata;
  f.p[0] = o.y - arc->o.y;
  f.p[1] = in == upper ? -1.0 : 1.0;
  f.p[2] = fabs(sr);
  f.p[3] = arc->o.x - o.x;
  f.in = !in;
  return f;
}

void Feature::rarp_endpoints (const Point &oa, const Point &p, const Angle &n,
			      const Point &op, bool flag, Points &res) const
{
  double d, alpha, e, beta, c[4], theta, psi, s = in ? r : - r;
  o.polar(d, alpha);
  Point q(s*n.c + p.x, s*n.s + p.y);
  q.polar(e, beta);
  if (!(flag ? circle_intersections(oa.x, oa.y, d, op.x, op.y, e, c, true) 
	: circle_intersections(op.x, op.y, e, oa.x, oa.y, d, c, true)))
    return;
  for (int i = 0; i < 2; ++i) {
    theta = mod_pi(atan2(c[2*i+1] - oa.y, c[2*i] - oa.x) - alpha);
    psi = mod_pi(atan2(c[2*i+1] - op.y, c[2*i] - op.x) - beta);
    res.push_back(flag ? Point(theta, psi) : Point(psi, theta));
  }
}

void Feature::rlrp_endpoints (const Point &ol, const Point &p, const Angle &n,
			      const Point &op, bool flag, Points &res) const
{
  double a, b, c, alpha, beta, st, ct, k, th[2];
  line_equation(a, b, c);
  alpha = atan2(b, a);
  beta = atan2(- n.s, - n.c);
  st = sin(alpha - beta);
  ct = cos(alpha - beta);
  k = c + a*(ct*p.x - st*p.y) + b*(st*p.x + ct*p.y);
  if (!sin_cos_roots(op.y - ol.y, op.x - ol.x, k, alpha, th))
    return;
  for (int i = 0; i < 2; ++i) {
    double v = mod_pi(th[i] + alpha - beta);
    res.push_back(flag ? Point(th[i], v) : Point(v, th[i]));
  }
}

void Feature::ratp_endpoints (const Point &p, const Angle &n, Points &res) 
  const
{
  double th[2], s = in ? r : - r;
  Point q(p.x + s*n.c, p.y + s*n.s);
  if (!sin_cos_roots(o.x, o.y, - q.y, 0.0, th))
    return;
  for (int i = 0; i < 2; ++i) {
    double x = cos(th[i])*o.x - sin(th[i])*o.y - q.x;
    res.push_back(Point(th[i], x));
  }
}

void Feature::rltp_endpoints (const Point &p, const Angle &n, Points &res) 
  const
{
  double a, b, c, alpha, beta, th, x, st, ct;
  line_equation(a, b, c);
  alpha = atan2(b, a);
  beta = atan2(- n.s, - n.c);
  th = mod_pi(beta - alpha);
  st = sin(th + alpha);
  ct = cos(th + alpha);
  if (ct == 0.0) return;
  x = - (c + p.x*ct + p.y*st)/ct;
  res.push_back(Point(th, x));
}

void Feature::rpta_endpoints (const Point &p, const Angle &n, Points &res) 
  const
{
  double th[2], s = in ? r : - r;
  Point q(p.x + s*n.c, p.y + s*n.s);
  if (!sin_cos_roots(q.x, q.y, - o.y, 0.0, th))
    return;
  for (int i = 0; i < 2; ++i) {
    double x = cos(th[i])*q.x - sin(th[i])*q.y - o.x;
    res.push_back(Point(th[i], x));
  }
}

void Feature::rptl_endpoints (const Point &p, const Angle &n, Points &res) 
  const
{
  double a, b, c, alpha, beta, st, ct, th, x;
  line_equation(a, b, c);
  if (a == 0.0) return;
  alpha = atan2(b, a);
  beta = atan2(- n.s, - n.c);
  th = mod_pi(alpha - beta);
  st = sin(th);
  ct = cos(th);
  x = (a*(ct*p.x - st*p.y) + b*(st*p.x + ct*p.y) + c)/a;
  res.push_back(Point(th, x));
}

bool Feature::contact (const Feature &b, bool aflag) const
{
  if (circle && b.circle)
    return contact_cc(b, aflag);
  else if (circle)
    return contact_cl(b, aflag);
  else if (b.circle)
    return b.contact_cl(*this, !aflag);

  return false;

}

bool Feature::contact_cc (const Feature &arc, bool aflag) const
{
  double sr = in ? r : - r, ss = arc.in ? arc.r : - arc.r;
  double d = sr + ss;
  Angle n(arc.o.x - o.x, arc.o.y - o.y, d), m(- n.c, - n.s, 1.0);
  Point p(o.x + sr*n.c, o.y + sr*n.s);
  return aflag ? arc.contact(p, m) : contact(p, n);
}

bool Feature::contact_cl (const Feature &line, bool aflag) const
{
  double a, b, c;
  line.line_equation(a, b, c);
  Point p(o.x - r*a, o.y - r*b);
  Angle n(- a, - b);
  return aflag ? line.contact(p, n) : contact(p, n);
}

bool Feature::contact (const Point &p, const Angle &n) const
{
  if (full_circle())
    return true;
  else if (circle) {
    double k1 = nt*n, k2 = n*nh, k3 = nt*nh;
    return k3 > 0.0 ? k1 >= 0.0 && k2 >= 0.0 : k1 >= 0.0 || k2 >= 0.0;
  }
  else {
    Point pt = p - t, ht = h - t;
    double k = pt%ht;
    return k >= 0.0 && k <= ht%ht;
  }
}

bool Feature::intersects (const Feature &b) const
{
  if (circle && b.circle)
    return intersects_cc(b);
  else if (circle)
    return intersects_cl(b);
  else if (b.circle)
    return b.intersects_cl(*this);
  else 
    return intersects_ll(b);
}

bool Feature::intersects_cc (const Feature &arc) const
{
  double xy[4];
  if (!circle_intersections(o.x, o.y, r, arc.o.x, arc.o.y, arc.r, xy))
    return false;
  for (int i = 0; i < 2; ++i) {
    Point p(xy[2*i], xy[2*i+1]);
    if (contains(p) && arc.contains(p))
      return true;
  }
  return false;
}

bool Feature::intersects_cl (const Feature &line) const
{
  double a, b, c, xy[4];
  line.line_equation(a, b, c);
  if (!circle_line_intersections(o.x, o.y, r, a, b, c, xy))
    return false;
  for (int i = 0; i < 2; ++i) {
    Point p(xy[2*i], xy[2*i+1]);
    if (contains(p) && line.contains(p))
      return true;
  }
  return false;
}

bool Feature::intersects_ll (const Feature &line) const
{
  double a1, b1, c1, a2, b2, c2, xy[2];
  a1 = t.y - h.y;
  b1 = h.x - t.x;
  c1 = - a1*h.x - b1*h.y;
  a2 = line.t.y - line.h.y;
  b2 = line.h.x - line.t.x;
  c2 = - a2*line.h.x - b2*line.h.y;
  if (!line_intersection(a1, b1, c1, a2, b2, c2, xy))
    return false;
  Point p(xy[0], xy[1]);
  return contains(p) && line.contains(p);
}

bool Feature::contains (const Point &p) const
{
  if (circle) {
    if (r > 0.0 && t == h)
      return true;
    Point d = in ? p - o : o - p;
    Angle n(d.x, d.y);
    double k1 = nt*n, k2 = n*nh;
    return nt*nh > 0.0 ? k1 >= 0.0 && k2 >= 0.0 : k1 >= 0.0 || k2 >= 0.0;
  }
  Point pt = p - t, ht = h - t;
  double k = pt%ht;
  return k >= 0.0 && k <= ht%ht;
}

bool Feature::above (const Point &p) const
{
  if (!circle) {
    bool inc = h.x > t.x;
    if (inc ? p.x <= t.x || p.x >= h.x : p.x <= h.x || p.x >= t.x)
      return false;
    return p.y < line_y(p.x, t, h);
  }
  if (p.x <= o.x - r || p.x >= o.x + r || p.y >= o.y + r) 
    return false;
  if (p.dist2(o) < r*r) {
    double s = in ? nt.angle() : nt.inv().angle();
    double e = in ? nh.angle() : nh.inv().angle();
    if (s == e)
      return true;
    if (s < 0.0)
      return e < s || e > 0.0 && p.x > h.x;
    else if (e > s)
      return p.x > h.x && p.x < t.x;
    else if (e < 0.0)
      return p.x < t.x;
    else
      return p.x < t.x || p.x > h.x;
  }
  else if (p.y < o.y)
    return t.x < p.x && h.x > p.x || t.x > p.x && h.x < p.x;
  else
    return false;
}

Feature * Feature::offset (double d) const
{
  Point ot(t.x + d*nt.c, t.y + d*nt.s), oh(h.x + d*nh.c, h.y + d*nh.s);
  if (!circle)
    return new Feature(ot, oh, in, nt, nh);
  else if (r > 0.0)
    return new Feature(ot, oh, o, in ? r + d : r - d, in, nt, nh);
  else if (d >= 0.0 ? in : !in)
    return new Feature(ot, oh, o, fabs(d), in, nt, nh);
  else
    return 0;
}

void Feature::discretize (double acc, Points &points) const
{
  if (points.empty())
    points.push_back(in ? t : h);
  if (!circle) {
    points.push_back(in ? h : t);
    return;
  }
  if (r == 0.0)
    return;
  double at, ah, ar, da, beta;
  at = in ? nt.angle() : mod_pi(nt.angle() + M_PI);
  ah = in ? nh.angle() : mod_pi(nh.angle() + M_PI);
  ar = at < ah ? ah - at : 2.0*M_PI + ah - at;
  da = acos(1.0 - acc/r);
  int n = (int) (ar/da);
  if (!in)
    da = - da;
  beta = in ?  at + da : ah + da;
  for (; n > 0; --n, beta += da)
    points.push_back(o + r*Point(cos(beta), sin(beta)));
  points.push_back(in ? h : t);
}

int Feature::split (Feature *parts) const
{
  if (full_circle()) {
    parts[0] = parts[1] = *this;
    parts[0].t = parts[1].h = Point(o.x + r, o.y);
    parts[0].nt = parts[1].nh = in ? 0.0 : M_PI;
    parts[0].h = parts[1].t = Point(o.x - r, o.y);
    parts[0].nh = parts[1].nt = in ? M_PI : 0.0;
    return 2;
  }
  Angle s = in ? nt : nt.inv(), e = in ? nh : nh.inv();
  if (s*e >= 0.0) {
    parts[0] = *this;
    return 1;
  }
  Angle m = s.midpoint(e);
  parts[0] = parts[1] = *this;
  parts[0].h = parts[1].t = o + r*m;
  parts[0].nh = parts[1].nt = in ? m : m.inv();
  return 2;
}

void Contact::rlta_set_in ()
{
  double u, v, w, alpha, th;
  a->line_equation(u, v, w, b->in);
  alpha = atan2(- v, - u);
  th = 0.5*(lb.x + ub.x);
  f.in = cos(th + alpha) > 0.0;
}

void Contact::discretize (double ytol, double xtol, bool rflag, 
			  Segments &res) const
{
  if (f.type == linear || lb.x == ub.x)
    res.push_back(new Segment(lb, ub, !f.in));
  else if (f.type == tata) {
    Point o(f.p[3], f.p[0]);
    double r = f.p[2];
    bool upper = f.p[1] > 0.0;
    res.push_back(new Segment(lb, ub, o, r, upper, !f.in));
  }
  else
    discretize_pline(ytol, xtol, rflag, res);
}

void Contact::discretize_pline (double ytol, double xtol, bool rflag, 
				Segments &res) const
{
  Points pline, st;
  pline.push_back(lb);
  st.push_back(ub);
  while (!st.empty()) {
    Point &p = *(pline.end()-1), &q = *(st.end()-1), m = 0.5*(p + q);
    if (p.x == m.x || q.x == m.x || p.x == q.x) {
      pline.push_back(q);
      st.pop_back();
    }
    else {
      double ym = f.y(m.x);
      if (fabs(ym - m.y) < ytol) {
	Point r = 0.5*(p + m);
	double yr = f.y(r.x);
	if (fabs(yr - r.y) < ytol) {
	  pline.push_back(q);
	  st.pop_back();
	}
	else {
	  st.push_back(Point(m.x, ym));
	  st.push_back(Point(r.x, yr));
	}
      }
      else if (rflag && q.x - p.x < xtol && fabs(q.y - p.y) > 6.2) {
	double s = p.y > 0.0 ? 1.0 : -1.0;
	pline.push_back(Point(m.x, s*M_PI));
	res.push_back(new Segment(pline, !f.in));
	pline.clear();
	pline.push_back(Point(m.x, -s*M_PI));
	pline.push_back(q);
	st.pop_back();
      }
      else
	st.push_back(Point(m.x, ym));
    }
  }
  res.push_back(new Segment(pline, !f.in));
}

void boundary_contacts (Boundary *a, bool ra, Boundary *b, bool rb, 
			const Point &ob, Contacts &res)
{
  for (unsigned int i = 0; i < a->size(); ++i)
    for (unsigned int j = 0; j < b->size(); ++j)
      ff_contacts(a->at(i), ra, b->at(j), rb, ob, res);
}

void ff_contacts (Feature *a, bool ra, Feature *b, bool rb, const Point &ob, 
		  Contacts &res)
{
  if (!a->compatible(b))
    return;
  if (ra && rb) {
    if (a->circle && b->circle)
      rara_contacts(a, b, ob, res);
    else if (a->circle)
      rarl_contacts(a, b, ob, res);
    else if (b->circle)
      rlra_contacts(a, b, ob, res);
    else
      rlrl_contacts(a, b, ob, res);
  }
  else if (ra) {
    if (a->circle && b->circle)
      rata_contacts(a, b, res);
    else if (a->circle)
      ratl_contacts(a, b, res);
    else if (b->circle)
      rlta_contacts(a, b, res);
    else
      rltl_contacts(a, b, res);
  }
  else if (a->circle && b->circle)
    tata_contacts(a, b, res);
  else if (a->circle)
    tatl_contacts(a, b, res);
  else if (b->circle)
    tlta_contacts(a, b, res);
}

void rara_contacts (Feature *a, Feature *b, const Point &ob, Contacts &res)
{
  ContactFunction f[2];
  int n = a->rara_function(b, ob, f);
  if (n == 0) return;
  Domain domain;
  f[0].rara_domain(domain);
  Points apoints, bpoints;
  rara_endpoints(a, b, ob, apoints, bpoints);
  contacts(a, b, n, f, domain, apoints, bpoints, res, &ob);
}

void rara_endpoints (Feature *a, Feature *b, const Point &ob, 
		     Points &apoints, Points &bpoints)
{
  Point oa(0.0, 0.0);
  if (!b->full_circle()) {
    a->rarp_endpoints(oa, b->t, b->nt, ob, true, apoints);
    a->rarp_endpoints(oa, b->h, b->nh, ob, true, apoints);
  }
  if (!a->full_circle()) {
    b->rarp_endpoints(ob, a->t, a->nt, oa, false, bpoints);
    b->rarp_endpoints(ob, a->h, a->nh, oa, false, bpoints);
  }
}  

void rarl_contacts (Feature *a, Feature *b, const Point &ob, Contacts &res)
{
  ContactFunction f[2];
  int n = a->rarl_function(b, ob, f);
  if (n == 0) return;
  Domain domain;
  f[0].rarl_domain(domain);
  Points apoints, bpoints;
  rarl_endpoints(a, b, ob, apoints, bpoints);
  contacts(a, b, n, f, domain, apoints, bpoints, res, &ob);
}

void rarl_endpoints (Feature *a, Feature *b, const Point &ob, 
		     Points &apoints, Points &bpoints)
{
  Point oa(0.0, 0.0);
  a->rarp_endpoints(oa, b->t, b->nt, ob, true, apoints);
  a->rarp_endpoints(oa, b->h, b->nh, ob, true, apoints);
  if (!a->full_circle()) {
    b->rlrp_endpoints(ob, a->t, a->nt, oa, false, bpoints);
    b->rlrp_endpoints(ob, a->h, a->nh, oa, false, bpoints);
  }
}

void rlra_contacts (Feature *a, Feature *b, const Point &ob, Contacts &res)
{
  ContactFunction f[2];
  int n = a->rlra_function(b, ob, f);
  if (n == 0) return;
  Domain domain;
  f[0].rlra_domain(domain);
  Points apoints, bpoints;
  rlra_endpoints(a, b, ob, apoints, bpoints, f[0].vertical());
  contacts(a, b, n, f, domain, apoints, bpoints, res, &ob);
}

void rlrl_contacts (Feature *a, Feature *b, const Point &ob, Contacts &res)
{
  Point oa;
  Points pt, ep;
  a->rlrp_endpoints(oa, b->t, b->nt, ob, true, pt);
  for (unsigned int i = 0; i < pt.size(); ++i) {
    Point mp = Angle(pt[i].y)*b->t + ob;
    if (a->move(pt[i].x, 0.0, 0.0).contains(mp))
      ep.push_back(pt[i]);
  }
  pt.clear();
  a->rlrp_endpoints(oa, b->h, b->nh, ob, true, pt);
  for (unsigned int i = 0; i < pt.size(); ++i) {
    Point mp = Angle(pt[i].y)*b->h + ob;
    if (a->move(pt[i].x, 0.0, 0.0).contains(mp))
      ep.push_back(pt[i]);
  }
  pt.clear();
  b->rlrp_endpoints(ob, a->t, a->nt, oa, false, pt);
  for (unsigned int i = 0; i < pt.size(); ++i)
    if (b->move(pt[i].y, ob.x, ob.y).contains(Angle(pt[i].x)*a->t))
      ep.push_back(pt[i]);
  pt.clear();
  b->rlrp_endpoints(ob, a->h, a->nh, oa, false, pt);
  for (unsigned int i = 0; i < pt.size(); ++i)
    if (b->move(pt[i].y, ob.x, ob.y).contains(Angle(pt[i].x)*a->h))
      ep.push_back(pt[i]);
  if (ep.size() < 2)
    return;
  ContactFunction f;
  f.type = linear;
  sort(ep.begin(), ep.end());
  for (unsigned int i = 0; i + 1 < ep.size(); ++i)
    if (ep[i] != ep[i+1])
      if (ep[i+1].x - ep[i].x > 0.5*M_PI) {
	Point p(- M_PI, ep[i].y), q(-p.x, p.y);
	res.push_back(new Contact(a, b, ep[i], p, f));
	res.push_back(new Contact(a, b, q, ep[i+1], f));
      }
      else if (fabs(ep[i].y - ep[i+1].y) > 0.5*M_PI) {
	Point p(ep[i].x, ep[i].y > 0.0 ? M_PI : - M_PI), q(p.x, -p.y);
	res.push_back(new Contact(a, b, ep[i], p, f));
	res.push_back(new Contact(a, b, q, ep[i+1], f));
      }
      else
	res.push_back(new Contact(a, b, ep[i], ep[i+1], f));
}

void rlra_endpoints (Feature *a, Feature *b, const Point &ob, 
		     Points &apoints, Points &bpoints, bool vflag)
{
  Point oa(0.0, 0.0);
  if (!b->full_circle()) {
    a->rlrp_endpoints(oa, b->t, b->nt, ob, true, apoints);
    a->rlrp_endpoints(oa, b->h, b->nh, ob, true, apoints);
  }
  if (vflag)
    return;
  b->rarp_endpoints(ob, a->t, a->nt, oa, false, bpoints);
  b->rarp_endpoints(ob, a->h, a->nh, oa, false, bpoints);
}

void rata_contacts (Feature *a, Feature *b, Contacts &res)
{
  ContactFunction f[2];
  int n = a->rata_function(b, f);
  if (n == 0) return;
  Domain domain;
  f[0].rata_domain(domain);
  Points apoints, bpoints;
  rata_endpoints(a, b, apoints, bpoints);
  contacts(a, b, n, f, domain, apoints, bpoints, res);
}

void rata_endpoints (Feature *a, Feature *b, Points &apoints,
		     Points &bpoints)
{
  if (!b->full_circle()) {
    a->ratp_endpoints(b->t, b->nt, apoints);
    a->ratp_endpoints(b->h, b->nh, apoints);
  }
  if (!a->full_circle()) {
    b->rpta_endpoints(a->t, a->nt, bpoints);
    b->rpta_endpoints(a->h, a->nh, bpoints);
  }
}

void ratl_contacts (Feature *a, Feature *b, Contacts &res)
{
  ContactFunction f[2];
  int n = a->ratl_function(b, f);
  if (n == 0) return;
  Domain domain;
  Points apoints, bpoints;
  ratl_endpoints(a, b, apoints, bpoints, f->vertical());
  contacts(a, b, n, f, domain, apoints, bpoints, res);
}

void ratl_endpoints (Feature *a, Feature *b, Points &apoints,
		     Points &bpoints, bool vflag)
{
  a->ratp_endpoints(b->t, b->nt, apoints);
  a->ratp_endpoints(b->h, b->nh, apoints);
  if (vflag || a->full_circle())
    return;
  b->rptl_endpoints(a->t, a->nt, bpoints);
  b->rptl_endpoints(a->h, a->nh, bpoints);
}

void rlta_contacts (Feature *a, Feature *b, Contacts &res)
{
  ContactFunction f[2];
  int n = a->rlta_function(b, f);
  if (n == 0) return;
  Domain domain;
  Points apoints, bpoints;
  rlta_endpoints(a, b, apoints, bpoints);
  Point p(mod_pi(- f->p[1]), 0.0), q(mod_pi(M_PI - f->p[1]), 0.0);
  if (!find_x(p.x, apoints))
    apoints.push_back(p);
  if (!find_x(q.x, apoints))
    apoints.push_back(q);
  int np = res.size();
  contacts(a, b, n, f, domain, apoints, bpoints, res);
  for (unsigned int i = np; i < res.size(); ++i)
    res[i]->rlta_set_in();
}

bool find_x (double x, const Points &points)
{
  for (unsigned int i = 0; i < points.size(); ++i)
    if (points[i].x == x)
      return true;
  return false;
}

void rlta_endpoints (Feature *a, Feature *b, Points &apoints,
		     Points &bpoints)
{
  if (!b->full_circle()) {
    a->rltp_endpoints(b->t, b->nt, apoints);
    a->rltp_endpoints(b->h, b->nh, apoints);
  }
  b->rpta_endpoints(a->t, a->nt, bpoints);
  b->rpta_endpoints(a->h, a->nh, bpoints);
}

void rltl_contacts (Feature *a, Feature *b, Contacts &res)
{
  Points pt, ep;
  a->rltp_endpoints(b->t, b->nt, pt);
  if (!pt.empty() &&
      a->move(pt[0].x, 0.0, 0.0).contains(b->t + Point(pt[0].y, 0.0)))
    ep.push_back(pt[0]);
  pt.clear();
  a->rltp_endpoints(b->h, b->nh, pt);
  if (!pt.empty() &&
      a->move(pt[0].x, 0.0, 0.0).contains(b->h + Point(pt[0].y, 0.0)))
    ep.push_back(pt[0]);
  pt.clear();
  b->rptl_endpoints(a->t, a->nt, pt);
  if (!pt.empty() &&
      b->move(0.0, pt[0].y, 0.0).contains(Angle(pt[0].x)*a->t))
    ep.push_back(pt[0]);
  pt.clear();
  b->rptl_endpoints(a->h, a->nh, pt);
  if (!pt.empty() && 
      b->move(0.0, pt[0].y, 0.0).contains(Angle(pt[0].x)*a->h))
    ep.push_back(pt[0]);
  if (ep.size() != 2 || ep[0] == ep[1])
    return;
  ContactFunction f;
  f.type = linear;
  sort(ep.begin(), ep.end());
  if (ep[1].x - ep[0].x > 0.5*M_PI) {
    Point p(- M_PI, ep[0].y), q(-p.x, p.y);
    res.push_back(new Contact(a, b, ep[0], p, f));
    res.push_back(new Contact(a, b, q, ep[1], f));
  }
  else
    res.push_back(new Contact(a, b, ep[0], ep[1], f));
}

void tata_contacts (Feature *a, Feature *b, Contacts &res)
{
  Angles aints, bints;
  a->split_vertical(aints);
  b->split_vertical(bints);
  for (unsigned int i = 0; i < aints.size(); i += 2)
    for (unsigned int j = 0; j < bints.size(); j += 2) {
      Angle s, e;
      if (intersect(aints[i].inv(), aints[i+1].inv(), bints[j], bints[j+1], 
		    s, e))
        res.push_back(tata_contact(a, b, s, e));
    }
}

bool intersect (const Angle &s1, const Angle &e1, const Angle &s2, 
		const Angle &e2, Angle &s, Angle &e)
{
  if (upper_arc(s1) != upper_arc(s2))
    return false;
  if (s1.in_arc(s2, e2)) {
    s = s1;
    e = e1.in_arc(s2, e2) ? e1 : e2;
    return true;
  }
  if (s2.in_arc(s1, e1)) {
    s = s2;
    e = e2.in_arc(s1, e1) ? e2 : e1;
    return true;
  }
  return false;
}

bool upper_arc (const Angle &s)
{
  return s.c == 1.0 && s.s == 0.0 || s.s > 0.0;
}

Contact * tata_contact (Feature *a, Feature *b, Angle s, Angle e)
{
  bool upper = upper_arc(s);
  ContactFunction f = a->tata_function(b, upper);
  Point p = b->point(s) - a->point(s.inv());
  Point q = b->point(e) - a->point(e.inv());
  Point pr(p.x, - p.y), qr(q.x, - q.y);
  if (pr < qr)
    return new Contact(a, b, pr, qr, f);
  else {
    f.in = !f.in;
    return new Contact(a, b, qr, pr, f);
  }
}

void tatl_contacts (Feature *a, Feature *b, Contacts &res)
{
  if (!b->nt.in_arc(a->nt.inv(), a->nh.inv()))
    return;
  Point pa = a->point(b->nt.inv()), p = b->h - pa, q = b->t - pa;
  bool in = a->in ? b->in : !b->in;
  Point pr(p.x, - p.y), qr(q.x, - q.y);
  if (pr < qr)
    res.push_back(new Contact(a, b, pr, qr, ContactFunction(in)));
  else
    res.push_back(new Contact(a, b, qr, pr, ContactFunction(!in)));
}

void tlta_contacts (Feature *a, Feature *b, Contacts &res)
{
  if (!a->nt.inv().in_arc(b->nt, b->nh))
    return;
  Point pb = b->point(a->nt.inv()), p = pb - a->t, q = pb - a->h;
  bool in = b->in ? a->in : !a->in;
  Point pr(p.x, - p.y), qr(q.x, - q.y);
  if (pr < qr)
    res.push_back(new Contact(a, b, pr, qr, ContactFunction(!in)));
  else
    res.push_back(new Contact(a, b, qr, pr, ContactFunction(in)));
}

void contacts (Feature *a, Feature *b, int n, const ContactFunction *f,
	       const Domain &domain, const Points &apoints, 
	       const Points &bpoints, Contacts &contacts, const Point *ob)
{
  Points agroups[6], bgroups[6];
  assign_groups(n, f, domain, apoints, agroups);
  assign_groups(n, f, domain, bpoints, bgroups);
  if (domain.empty()) {
    ContactIntervals aints, bints;
    form_intervals1(a, b, ob, n, f, agroups, true, aints);
    form_intervals1(a, b, ob, n, f, bgroups, false, bints);
    intersect_intervals1(a, b, n, f, aints, bints, contacts);
  }
  else
    for (unsigned int i = 0; i < domain.size(); i += 2) {
      Domain di;
      di.push_back(domain[i]);
      di.push_back(domain[i+1]);
      ContactFunction nf[2] = {f[0], f[1]};
      double xt = di[0] + max(0.0001, 0.5*(di[1] - di[0]));
      if (f[0].y(xt) > f[1].y(xt)) {
	reverse(nf, nf + 2);
	reverse(agroups + i, agroups + i + 2);
	reverse(bgroups + i, bgroups + i + 2);
      }
      ContactIntervals aints, bints;
      form_intervals2(a, b, ob, nf, di, agroups + i, true, aints);
      form_intervals2(a, b, ob, nf, di, bgroups + i, false, bints);
      intersect_intervals2(a, b, n, nf, di, aints, bints, contacts);
    }
}

void assign_groups (int n, const ContactFunction *f, const Domain &domain, 
		    const Points &points, Points *groups)
{
  for (unsigned int k = 0; k < points.size(); ++k) {
    const Point &p = points[k];
    if (domain.size() == 4 && domain[1] == domain[2] && p.x == domain[1]) {
      add_point(p, groups[0], 0, f, domain);
      add_point(p, groups[2], 2, f, domain);
      continue;
    }
    int i = 0, j = 0;
    if (domain.size() == 4 &&
	(in_arc(p.x, domain[2], domain[3]) ||
	 min(fabs(p.x - domain[2]), fabs(p.x - domain[3])) <
	 min(fabs(p.x - domain[0]), fabs(p.x - domain[1]))))
      i = 1;
    if (f->vertical()) {
      if (fabs(p.x - f[1].y(p.x)) < fabs(p.x - f[0].y(p.x)))
	j = 1;
    }
    else if (n == 2) {
      bool ytrans = f->type == rata;
      double d0 = ytrans ? p.y - f[0].y(p.x) : mod_pi(p.y - f[0].y(p.x));
      double d1 = ytrans ? p.y - f[1].y(p.x) : mod_pi(p.y - f[1].y(p.x));
      if (fabs(d1) < fabs(d0))
	j = 1;
    }
    add_point(p, groups[2*i+j], i, f, domain);
  }
}

void add_point (const Point &p, Points &points, int i, 
		const ContactFunction *f, const Domain &domain)
{
  int n = points.size();
  if (n == 0) {
    points.push_back(p);
    return;
  }
  if (f->vertical()) {
    if (p.y < points[0].y)
      points.insert(points.begin(), p);
    else if (p.y > points[n-1].y)
      points.push_back(p);
    else
      for (int j = 1; j < n; ++j)
        if (in_arc(p.y, points[j-1].y, points[j].y)) {
          if (!(p == points[j-1] || p == points[j]))
            points.insert(points.begin() + j, p);
          return;
        }
  }
  else {
    double x1, ub;
    if (domain.empty() || domain[2*i] < domain[2*i+1]) {
      x1 = - M_PI;
      ub = M_PI;
    }
    else 
      x1 = ub = 0.5*(domain[2*i] + domain[2*i+1]);
    for (int j = 0; j <= n; ++j) {
      double x2 = j < n ? points[j].x : ub;
      if (in_arc(p.x, x1, x2)) {
        if ((j == 0 || p != points[j-1]) && (j == n || p != points[j]))
          points.insert(points.begin() + j, p);
        return;
      }
      x1 = x2;
    }
  }
}

void form_intervals1 (Feature *a, Feature *b, const Point *ob, int n,
		      const ContactFunction *f, Points *groups, bool aflag, 
		      ContactIntervals &ints)
{
  for (int i = 0; i < n; ++i) {
    const Points &group = groups[i];
    if (f[i].vertical() && (!aflag || group.empty())) {
      if (contact_occurs(a, b, ob, f[i].y(0.0), 0.0, false)) {
	Point lb(f[i].y(0.0), ob ? - M_PI : -1e10), ub(lb.x, - lb.y);
	ints.push_back(ContactInterval(lb, ub, i == 0));
      }
    }
    else if (!f[i].vertical() && group.size() < 2) {
      Point lb(- M_PI, f[i].y(- M_PI)), ub(M_PI, lb.y);
      if (contact_occurs1(a, b, ob, f[i], lb, ub, aflag))
	ints.push_back(ContactInterval(lb, ub, i == 0));
    }
    else if (f->vertical() && !ob && a->circle && !b->circle) {
      for (unsigned int j = 0; j + 1 < group.size(); ++j)
	if (contact_occurs1(a, b, ob, f[i], group[j], group[j+1], aflag))
	  ints.push_back(ContactInterval(group[j], group[j+1], i == 0));
    }
    else {
      int m = group.size();
      for (int j = 0; j < m; ++j) {
	int k = j == m - 1 ? 0 : j + 1;
	if (contact_occurs1(a, b, ob, f[i], group[j], group[k], aflag))
	  ints.push_back(ContactInterval(group[j], group[k], i == 0));
      }
    }
  }
}

bool contact_occurs1 (Feature *a, Feature *b, const Point *ob, 
		      const ContactFunction &f, const Point &lb, 
		      const Point &ub, bool aflag)
{
  if (lb == ub) return false;
  double u, v;
  if (f.type == linear) {
    u = 0.5*(lb.x + ub.x);
    v = lb.y < ub.y ? 0.5*(lb.y + ub.y) : mod_pi(0.5*(lb.y + ub.y) + M_PI);
  }
  else {
    u = lb.x < ub.x ? 0.5*(lb.x + ub.x) : mod_pi(0.5*(lb.x + ub.x) + M_PI);
    bool flag = true;
    v = f.y(u, flag);
    if (!flag)
      return false;
  }
  return contact_occurs(a, b, ob, u, v, aflag);
}

void intersect_intervals1 
(Feature *a, Feature *b, int n, const ContactFunction *f, 
 const ContactIntervals &aints, const ContactIntervals &bints,
 Contacts &contacts)
{
  for (unsigned int i = 0; i < aints.size(); ++i)
    for (unsigned int j = 0; j < bints.size(); ++j) {
      ContactIntervals ints;
      aints[i].intersect(bints[j], f, ints);
      for (unsigned int k = 0; k < ints.size(); ++k) {
	const ContactFunction &fk = ints[k].tlower ? f[0] : f[1];
	contacts.push_back(new Contact(a, b, ints[k].t, ints[k].h, fk));
      }
    }
}

void form_intervals2 (Feature *a, Feature *b, const Point *ob,
		      const ContactFunction *f, const Domain &domain,
		      Points *groups, bool aflag, ContactIntervals &ints)
{
  if (groups[0].size() + groups[1].size() < 2) {
    Point lb = domain_point(f, domain[0]), ub = domain_point(f, domain[1]);
    if (contact_occurs1(a, b, ob, f[0], lb, ub, aflag))
      ints.push_back(ContactInterval(lb, true, lb, false));
    return;
  }
  reverse(groups[1].begin(), groups[1].end());
  int nl = groups[0].size(), nu = groups[1].size(), n = nl + nu;
  for (int i = 0; i < n; ++i) {
    int j = i < n - 1 ? i + 1 : 0;
    bool tlower = i < nl, hlower = j < nl;
    const Point &t = tlower ? groups[0][i] : groups[1][i-nl];
    const Point &h = hlower ? groups[0][j] : groups[1][j-nl];
    ContactInterval cint(t, tlower, h, hlower);
    if (contact_occurs2(a, b, ob, f, domain, aflag, cint))
      ints.push_back(cint);
  }
}

bool contact_occurs2 (Feature *a, Feature *b, const Point *ob, 
		      const ContactFunction *f, const Domain &domain,
		      bool aflag, const ContactInterval &cint)
{
  bool tlower = cint.tlower, hlower = cint.hlower;
  double lb = domain[0], ub = domain[1], tx = cint.t.x, hx = cint.h.x, u, v;  
  if (ub < lb) {
    if (tx < 0.5*(lb+ub))
      tx += 2.0*M_PI;
    if (hx < 0.5*(lb+ub))
      hx += 2.0*M_PI;
    ub += 2.0*M_PI;
  }
  if (tlower && hlower && tx < hx) {
    u = 0.5*(tx + hx);
    v = f[0].y(u);
  }
  else if (tlower && hlower) {
    u = 0.5*(lb + ub);
    v = f[1].y(u);
  }
  else if (!tlower && !hlower && tx > hx) {
    u = 0.5*(tx + hx);
    v = f[1].y(u);
  }
  else if (!tlower && !hlower) {
    u = 0.5*(lb + ub);
    v = f[0].y(u);
  }
  else if (tlower) {
    double dt = ub - tx, dh = ub - hx, d = 0.5*(dt + dh);
    if (dt < dh) {
      u = hx + d;
      v = f[1].y(u);
    }
    else {
      u = tx + d;
      v = f[0].y(u);
    }
  }
  else {
    double dt = tx - lb, dh = hx - lb, d = 0.5*(dt + dh);
    if (dt < dh) {
      u = hx - d;
      v = f[0].y(u);
    }
    else {
      u = tx - d;
      v = f[1].y(u);
    }
  }
  return contact_occurs(a, b, ob, u, v, aflag);
}

void intersect_intervals2
(Feature *a, Feature *b, int n, const ContactFunction *f, 
 const Domain &domain, const ContactIntervals &aints, 
 const ContactIntervals &bints, Contacts &contacts)
{
  for (unsigned int i = 0; i < aints.size(); ++i)
    for (unsigned int j = 0; j < bints.size(); ++j) {
      ContactIntervals ints;
      aints[i].intersect(bints[j], f, domain, ints);
      for (unsigned int k = 0; k < ints.size(); ++k) {
	const ContactFunction &fk = ints[k].tlower ? f[0] : f[1];
	contacts.push_back(new Contact(a, b, ints[k].t, ints[k].h, fk));
      }
    }
}

bool contact_occurs (Feature *a, Feature *b, const Point *ob, 
		     double u, double v, bool aflag)
{
  Feature ma = a->move(u, 0.0, 0.0);
  Feature mb = ob ? b->move(v, ob->x, ob->y) : b->move(0.0, v, 0.0);
  return ma.contact(mb, aflag);
}

bool sin_roots (double a, double offset, double *th, double tol)
{
  double as;
  bool flag = true;
  if (a >= 1.0) {
    as = 0.5*M_PI;
    flag = a < 1.0 + tol;
  }
  else if (a <= -1.0) {
    as = - 0.5*M_PI;
    flag = a > -1.0 - tol;
  }
  else
    as = asin(a);
  th[0] = mod_pi(as - offset);
  th[1] = mod_pi(M_PI - as - offset);
  return flag;
}

bool cos_roots (double a, double offset, double *th)
{
  if (fabs(a) >= 1.0) 
    return false;
  th[0] = mod_pi(acos(a) - offset);
  th[1] = mod_pi(- acos(a) - offset);
  return true;
}

bool sin_cos_roots (double a, double b, double c, double offset, double *th,
		    double tol)
{
  if (a == 0.0 && b == 0.0)
    return false;
  double g, gamma;
  Point ab(a, b);
  ab.polar(g, gamma);
  return sin_roots(- c/g, offset + gamma, th, tol);
}

double mod_pi (double alpha, bool flag)
{
  static double k = 2.0*M_PI;
  int d = (int) (alpha/k);
  double y = alpha - d*k;
  if (y > M_PI)
    y -= k;
  else if (y < - M_PI)
    y += k;
  if (flag && y == M_PI)
    y = - M_PI;
  return y;
}

bool in_arc (double x, double lb, double ub)
{
  return lb <= ub ? x >= lb && x <= ub : x >= lb || x <= ub;
}

Boundary * fillet (const Boundary &fin)
{
  Feature *fp = fin[0];
  Boundary *boundary = new Boundary;
  boundary->push_back(fp);
  int n = fin.size();
  if (n == 1 || n == 2 && fp->merge(fin[1])) {
    fp->h = fp->t;
    fp->nh = fp->nt;
    if (n == 2) 
      delete fin[1];
    return boundary;
  }
  for (int i = 1; i < n; ++i) 
    if (fp->merge(fin[i]))
      delete fin[i];
    else {
      fillet(fp, fin[i], *boundary);
      boundary->push_back(fin[i]);
      fp = fin[i];
    }
  if (boundary->at(0)->merge(fp)) {
    boundary->pop_back();
    delete fp;
  }
  else
    fillet(fp, boundary->at(0), *boundary);
  return boundary;
}

void fillet (Feature *a, Feature *b, Boundary &boundary)
{
  Point &pa = a->in ? a->h : a->t, &pb = b->in ? b->t : b->h;
  pa = pb;
  Angle &na = a->in ? a->nh : a->nt, &nb = b->in ? b->nt : b->nh;
  double k = na*nb;
  if (fabs(k) < 1e-10)
    if (a->circle)
      na = nb;
    else
      nb = na;
  else if (k > 0.0)
    boundary.push_back(new Feature(pb, na, nb, true));
  else
    boundary.push_back(new Feature(pb, nb, na, false));
}

Boundary * offset (Boundary *boundary, double d)
{
  Boundary b;
  for (unsigned int i = 0; i < boundary->size(); ++i) {
    Feature *f = boundary->at(i)->offset(d);
    if (f)
      b.push_back(f);
  }
  int n = b.size(), np = n - 1;
  for (int i = 0; i < n; ++i) {
    b[np]->trim(b[i]);
    np = i;
  }
  Boundary *res = fillet(b);
  return res;
}

void delete_boundary (Boundary *boundary)
{
  for (unsigned int i = 0; i < boundary->size(); ++i)
    delete boundary->at(i);
  delete boundary;
}

void form_curves (bool arot, bool brot, const Contacts &contacts, double acc,
		  double *bbox, Segments &curves)
{
  for (unsigned int i = 0; i < contacts.size(); ++i)
    contacts[i]->discretize(acc, acc, brot, curves);
  form_curves_bbox(brot, bbox, curves);
}

void form_curves_bbox (bool brot, double *bbox, Segments &curves)
{
  Point bl(bbox[0], bbox[2]), br(bbox[1], bbox[2]), tr(bbox[1], bbox[3]),
    tl(bbox[0], bbox[3]);
  curves.push_back(new Segment(br, tr, true));
  curves.push_back(new Segment(bl, tl, false));
  if (brot)
    form_bbox_y(curves, bbox);
  else {
    curves.push_back(new Segment(bl, br, true));
    curves.push_back(new Segment(tl, tr, false));
  }
}

void form_bbox_y (Segments &curves, double *bbox)
{
  vector<double> xl, xu;
  for (unsigned int i = 0; i < curves.size(); ++i) {
    Point p = curves[i]->tail(), q = curves[i]->head();
    if (p.y == bbox[2])
      xl.push_back(p.x);
    else if (p.y == bbox[3])
      xu.push_back(p.x);
    if (q.y == bbox[2])
      xl.push_back(q.x);
    else if (q.y == bbox[3])
      xu.push_back(q.x);
  }
  form_bbox_y_aux(curves, bbox, bbox[2], true, xl);
  form_bbox_y_aux(curves, bbox, bbox[3], false, xu);
}

void form_bbox_y_aux (Segments &curves, double *bbox, double y, bool in,
		      vector<double> &x)
{
  x.push_back(bbox[0]);
  x.push_back(bbox[1]);
  sort(x.begin(), x.end());
  for (unsigned int i = 0; i + 1 < x.size(); ++i)
    if (x[i] < x[i+1])
      curves.push_back(new Segment(Point(x[i], y), Point(x[i+1], y), in));
}

void write_boundary (const Boundary &boundary, ostream &ostr)
{
  ostr << "(";
  for (unsigned int i = 0; i < boundary.size(); ++i)
    write_feature(boundary[i], ostr);
  ostr << ")" << endl;
}

void write_feature (Feature *f, ostream &ostr)
{
  ostr << setprecision(16) << "(";
  write_point(f->t, ostr);
  ostr << " ";
  write_point(f->h, ostr);
  ostr << " " << f->in << " " << f->circle << " ";
  if (f->circle) {
    write_point(f->o, ostr);
    ostr << " " << f->r << " " << f->nt.angle() << " " << f->nh.angle();
  }
  ostr << ")";
}

void write_point (const Point &p, ostream &ostr)
{
  ostr << "(" << p.x << " " << p.y << ")";
}
