#ifndef CS
#define CS

#include "arrange.h"

#define EPSILON 1e-14
#define BIGNUMBER 1e20

typedef vector<double> Domain;

enum ContactType { ratl, rlta, rata, rarl, rlra, rara, tata, linear };

class ContactFunction {
 public:
  ContactFunction () {}
  ContactFunction (bool iin) : type(linear), in(iin) {}
  bool vertical () const { return type == linear && p[1] == 0.0; }
  double y (double x) const { bool flag = true; return y(x, flag); }
  double y (double x, bool &flag) const;
  double y_rr (double x, bool &flag) const;
  double y_rata (double x, bool &flag) const;
  double y_rlta (double x, bool &flag) const;
  double y_tata (double x, bool &flag) const;
  void rara_domain (Domain &res) const;
  void rarl_domain (Domain &res) const;
  void rlra_domain (Domain &res) const;
  void rata_domain (Domain &res) const;
  void domain_aux (int n, double *x, Domain &res) const;

  ContactType type;
  double p[10];
  bool in;
};

Point domain_point (const ContactFunction *f, double x);

bool domain_order (double x1, double x2, const Domain &domain);

class ContactInterval {
 public:
  ContactInterval () {}
  ContactInterval (const Point &it, bool itlower, const Point &ih, bool ihlower)
    : t(it), h(ih), tlower(itlower), hlower(ihlower) {}
  ContactInterval (const Point &it, const Point &ih, bool lower)
    : t(it), h(ih), tlower(lower), hlower(lower) {}
  void intersect (const ContactInterval &bint, const ContactFunction *f, 
		  vector<ContactInterval> &ints) const;
  void intersect_g (const ContactInterval &bint, const ContactFunction *f,
		    vector<ContactInterval> &ints) const;
  void simple_intervals_v (const ContactFunction *f,
			   vector<ContactInterval> &ints) const;
  void intersect_v (const ContactInterval &bint, const ContactFunction *f,
		    vector<ContactInterval> &ints) const;
  void simple_intervals_g (const ContactFunction *f, 
			   vector<ContactInterval> &ints) const;
  void intersect (const ContactInterval &bint, const ContactFunction *f, 
		  const Domain &domain, vector<ContactInterval> &ints) const;
  bool contains (const Point &p, bool plower, const Domain &domain) const;
  void simple_intervals (const ContactFunction *f, const Domain &domain,
			 vector<ContactInterval> &ints) const;

  Point t, h;
  bool tlower, hlower;
};

typedef vector<ContactInterval> ContactIntervals;

void intersect_ep (const Point &a, bool alower, const Point &b, bool blower,
		   ContactIntervals &ints);

ContactInterval simple_interval (const Point &t, const Point &h, 
				 bool lower, const Domain &domain);

class Feature {
 public:
  Feature () {}
  Feature (const Point &it, const Point &ih, bool iin)
    : circle(false), in(iin), t(it), h(ih) { set_normals(); }
  Feature (const Point &it, const Point &ih, const Point &io,
	   double ir, bool iin)
    : circle(true), in(iin), t(it), h(ih), o(io), r(ir) { set_normals(); }
  Feature (const Point &it, const Angle &nti, const Angle &nhi, bool iin)
    : circle(true), in(iin), t(it), h(it), o(it), r(0.0), nt(nti), nh(nhi) {}
  Feature (const Point &it, const Point &ih, bool iin, const Angle &nti, 
	   const Angle &nhi) 
    : circle(false), in(iin), t(it), h(ih), nt(nti), nh(nhi) {}
  Feature (const Point &it, const Point &ih, const Point &io,
	   double ir, bool iin, const Angle &nti, const Angle &nhi)
    : circle(true), in(iin), t(it), h(ih), o(io), r(ir), nt(nti), nh(nhi) {}
  void set_normals ();
  bool full_circle () const { return circle && nt == nh; }
  Point point (const Angle &th) const;
  void split_vertical (Angles &ints) const;
  bool merge (Feature *f);
  void trim (Feature *f); 
  Point closest_point (Feature *f, const Point &p) const;
  void line_equation (double &a, double &b, double &c, bool flag = true) const;
  Feature move (double th, double x, double y) const;
  Feature move (const Angle &a, const Point &d) const;
  bool compatible (Feature *b) const;
  double aa_r (const Feature &arc) const;
  int rara_function (Feature *arc, const Point &q, ContactFunction *f) const;
  int rarl_function (Feature *line, const Point &q, ContactFunction *f) const;
  int rlra_function (Feature *arc, const Point &q, ContactFunction *f) const;
  int rata_function (Feature *arc, ContactFunction *f) const;
  int ratl_function (Feature *line, ContactFunction *f) const;
  int rlta_function (Feature *arc, ContactFunction *f) const;
  ContactFunction tata_function (Feature *arc, bool upper) const;
  void rarp_endpoints (const Point &oa, const Point &p, const Angle &n,
		       const Point &op, bool flag, Points &res) const;
  void rlrp_endpoints (const Point &ol, const Point &p, const Angle &n,
		       const Point &op, bool rflag, Points &res) const;
  void ratp_endpoints (const Point &p, const Angle &n, Points &res) const;
  void rltp_endpoints (const Point &p, const Angle &n, Points &res) const;
  void rpta_endpoints (const Point &p, const Angle &n, Points &res) const;
  void rptl_endpoints (const Point &p, const Angle &n, Points &res) const;
  bool contact (const Feature &b, bool aflag) const;
  bool contact_cc (const Feature &arc, bool aflag) const;
  bool contact_cl (const Feature &line, bool aflag) const;
  bool contact (const Point &p, const Angle &n) const;
  bool intersects (const Feature &b) const;
  bool intersects_cc (const Feature &arc) const;
  bool intersects_cl (const Feature &line) const;
  bool intersects_ll (const Feature &line) const;
  bool contains (const Point &p) const;
  bool above (const Point &p) const;
  Feature * offset (double d) const;
  void discretize (double acc, Points &points) const;
  int split (Feature *parts) const;

  bool circle, in;
  Point t, h, o;
  double r;
  Angle nt, nh;
};

typedef vector<Feature *> Boundary;

class Contact {
 public:
  Contact (Feature *ia, Feature *ib, const Point &ilb, const Point &iub,
	   const ContactFunction &fi) 
    : a(ia), b(ib), lb(ilb), ub(iub), f(fi) {}
  void rlta_set_in ();
  void discretize (double ytol, double xtol, bool rflag, 
		   Segments &res) const;
  void discretize_pline (double ytol, double xtol, bool rflag, 
			 Segments &res) const;

  Feature *a, *b;
  Point lb, ub;
  ContactFunction f;
};

typedef vector<Contact *> Contacts;

void boundary_contacts (Boundary *a, bool ra, Boundary *b, bool rb,
			const Point &ob, Contacts &res);

void ff_contacts (Feature *a, bool ra, Feature *b, bool rb, const Point &ob, 
		  Contacts &res);

void rara_contacts (Feature *a, Feature *b, const Point &ob, Contacts &res);

void rara_endpoints (Feature *a, Feature *b, const Point &ob, 
		     Points &apoints, Points &bpoints);

void rarl_contacts (Feature *a, Feature *b, const Point &ob, Contacts &res);

void rarl_endpoints (Feature *a, Feature *b, const Point &ob, 
		     Points &apoints, Points &bpoints);

void rlra_contacts (Feature *a, Feature *b, const Point &ob, Contacts &res);

void rlra_endpoints (Feature *a, Feature *b, const Point &ob, 
		     Points &apoints, Points &bpoints, bool vflag);

void rlrl_contacts (Feature *a, Feature *b, const Point &ob, Contacts &res);

void rata_contacts (Feature *a, Feature *b, Contacts &res);

void rata_endpoints (Feature *a, Feature *b, Points &apoints,
		     Points &bpoints);

void ratl_contacts (Feature *a, Feature *b, Contacts &res);

void ratl_endpoints (Feature *a, Feature *b, Points &apoints,
		     Points &bpoints, bool vflag);

void rlta_contacts (Feature *a, Feature *b, Contacts &res);

bool find_x (double x, const Points &points);

void rlta_endpoints (Feature *a, Feature *b, Points &apoints,
		     Points &bpoints);

void rltl_contacts (Feature *a, Feature *b, Contacts &res);

void tata_contacts (Feature *a, Feature *b, Contacts &res);

bool intersect (const Angle &s1, const Angle &e1, const Angle &s2, 
		const Angle &e2, Angle &s, Angle &e);

bool upper_arc (const Angle &s);

Contact * tata_contact (Feature *a, Feature *b, Angle s, Angle e);

void tatl_contacts (Feature *a, Feature *b, Contacts &res);

void tlta_contacts (Feature *a, Feature *b, Contacts &res);

void contacts (Feature *a, Feature *b, int n, const ContactFunction *f,
	       const Domain &domain, const Points &apoints, 
	       const Points &bpoints, Contacts &contacts, const Point *ob = 0);

void assign_groups (int n, const ContactFunction *f, const Domain &domain, 
		    const Points &points, Points *groups);

void add_point (const Point &p, Points &points, int i, 
		const ContactFunction *f, const Domain &domain);

void form_intervals1 (Feature *a, Feature *b, const Point *ob, int n,
		      const ContactFunction *f, Points *groups, bool aflag, 
		      ContactIntervals &ints);

bool contact_occurs1 (Feature *a, Feature *b, const Point *ob, 
		      const ContactFunction &f, const Point &lb, 
		      const Point &ub, bool aflag);

void intersect_intervals1 (Feature *a, Feature *b, int n, const ContactFunction *f, 
			   const ContactIntervals &aints,
			   const ContactIntervals &bints, Contacts &contacts);

void form_intervals2 (Feature *a, Feature *b, const Point *ob,
		      const ContactFunction *f, const Domain &domain,
		      Points *groups, bool aflag, ContactIntervals &ints);

bool contact_occurs2 (Feature *a, Feature *b, const Point *ob, 
		      const ContactFunction *f, const Domain &domain,
		      bool aflag, const ContactInterval &cint);

void intersect_intervals2 (Feature *a, Feature *b, int n, const ContactFunction *f, 
			   const Domain &domain, const ContactIntervals &aints,
			   const ContactIntervals &bints, Contacts &contacts);

bool contact_occurs (Feature *a, Feature *b, const Point *ob, 
		     double u, double v, bool aflag);

bool sin_roots (double a, double offset, double *th, double tol = 0.0);

bool cos_roots (double a, double offset, double *th);

bool sin_cos_roots (double a, double b, double c, double offset, 
		    double *th1, double tol = 0.0);

double mod_pi (double alpha, bool flag = false);

bool in_arc (double x, double lb, double ub);

Boundary * fillet (const Boundary &fin);

void fillet (Feature *a, Feature *b, Boundary &boundary);

Boundary * offset (Boundary *boundary, double d);

void delete_boundary (Boundary *boundary);

void form_curves (bool arot, bool brot, const Contacts &contacts, double acc, 
		  double *bbox, Segments &curves);

void form_curves_bbox (bool brot, double *bbox, Segments &curves);

void form_bbox_y (Segments &curves, double *bbox);
 
void form_bbox_y_aux (Segments &curves, double *bbox, double x, bool in, 
		      vector<double> &y);

void write_boundary (const Boundary &boundary, ostream &ostr);

void write_feature (Feature *f, ostream &ostr);

void write_point (const Point &p, ostream &ostr);

#endif

