#include "mtriangulate.h"

PVertexType PVertex::type () const
{
  PVertex *prev = edge->next->twin->tail, *next = edge->twin->tail;
  bool pflag = *prev < *this, nflag = *this < *next;
  if (pflag && nflag)
    return Right;
  if (!pflag && !nflag)
    return Left;
  bool rflag = signedArea(prev->p, p, next->p) > 0.0;
  if (pflag)
    return rflag ? Start : Split;
  return rflag ? End : Merge;
}

void PVertex::link (PVertex *w)
{
  PEdge *e = new PEdge(this), *f = new PEdge(w, e);
  e->twin = f;
  insert(e);
  w->insert(f);
}

void PVertex::insert (PEdge *e)
{
  Angle ae(e->twin->tail->p - p);
  PEdge *f = edge;
  Angle af(f->twin->tail->p - p);
  while (true) {
    PEdge *g = f->next;
    Angle ag(g->twin->tail->p - p);
    if (g == edge || ag != af && ae.in_arc(ag, af)) {
      f->next = e;
      e->next = g;
      return;
    }
    f = g;
    af = ag;
  }
}

bool PEdgeOrder::operator() (PEdge *a, PEdge *b) const
{ 
  return a->in ? a->left(b->tail->p) : !b->left(a->tail->p);
}

bool PEdge::left (const Point &p) const
{
  const Point &t = tail->p, &h = twin->tail->p;
  if (t.x == h.x)
    return t.x < p.x;
  double x = t.x + (p.y - t.y)*(h.x - t.x)/(h.y - t.y);
  return x < p.x;
}

void PEdge::monotonic_loop (Points &pl, Points &pr)
{
  PEdge *e = twin->next;
  while (true) {
    pl.push_back(e->tail->p);
    PEdge *f = e->twin->next;
    if (*e->tail < *f->tail)
      break;
    e->flag = true;
    e = f;
  }
  while (!e->flag) {
    pr.push_back(e->tail->p);
    e->flag = true;
    e = e->twin->next;
  }
  reverse(pr.begin(), pr.end());
}

void triangulate (const Profile &region, Triangles &triangles)
{
  PVertices vertices;
  form_loops(region, vertices);
  if (vertices.size() > 3)
    make_y_monotonic(vertices);
  if (vertices.size() > 2)
    triangulate_monotonic(vertices, triangles);
  delete_loop(vertices);
}

void form_loops (const Profile &region, PVertices &vertices)
{
  for (unsigned int i = 0; i < region.size(); ++i)
    form_loop(region[i], vertices);
}

void form_loop (const Points &polygon, PVertices &vertices)
{
  int n = vertices.size(), np = n, m = polygon.size();
  if (m < 3) return;
  for (int i = 0; i < m; ++i)
    if (i == 0 || 
	vertices[n-1]->p.dist2(polygon[i]) > 1e-12 &&
	polygon[0] != polygon[i]) {
      PVertex *v = new PVertex(polygon[i]);
      v->edge = new PEdge(v);
      v->edge->next = new PEdge(v, 0, v->edge, true);
      vertices.push_back(v);
      ++n;
    }
  int ip = n - 1;
  for (int i = np; i < n; ++i) {
    int in = i < n - 1 ? i + 1 : np;
    vertices[i]->edge->twin = vertices[in]->edge->next;
    vertices[i]->edge->next->twin = vertices[ip]->edge;
    ip = i;
  }
}

void make_y_monotonic (PVertices &vertices)
{
  sort(vertices.begin(), vertices.end(), vertex_order);
  PEdgeSet sweep;
  PVertices diagonals;
  for (unsigned int i = 0; i < vertices.size(); ++i) {
    PVertex *v = vertices[i];
    switch (v->type()) {
    case Start:
      handle_start(v, sweep);
      break;
    case End:
      handle_end(v, sweep, diagonals);
      break;
    case Left:
      handle_left(v, sweep, diagonals);
      break;
    case Right:
      handle_right(v, sweep, diagonals);
      break;
    case Merge:
      handle_merge(v, sweep, diagonals);
      break;
    case Split:
      handle_split(v, sweep, diagonals);
    }
  }
  for (unsigned int i = 0; i < diagonals.size(); i += 2)
    diagonals[i]->link(diagonals[i+1]);
}

void handle_start (PVertex *v, PEdgeSet &sweep)
{
  PEdge *e = v->edge;
  e->node = sweep.insert(e).first;
  e->in = true;
  e->helper = v;
}

void handle_end (PVertex *v, PEdgeSet &sweep, PVertices &diagonals)
{
  PEdge *d = v->edge->next->twin;
  check_diagonal(v, d->helper, diagonals);
  sweep.erase(d->node);
}

void handle_left (PVertex *v, PEdgeSet &sweep, PVertices &diagonals)
{
  PEdge *d = v->edge->next->twin, *e = v->edge;
  check_diagonal(v, d->helper, diagonals);
  sweep.erase(d->node);
  e->node = sweep.insert(e).first;
  e->in = true;
  e->helper = v;
}

void handle_right (PVertex *v, PEdgeSet &sweep, PVertices &diagonals)
{
  PEdge *f = find_left(v, sweep);
  check_diagonal(v, f->helper, diagonals);
  f->helper = v;
}

void handle_merge (PVertex *v, PEdgeSet &sweep, PVertices &diagonals)
{
  PEdge *d = v->edge->next->twin;
  check_diagonal(v, d->helper, diagonals);
  sweep.erase(d->node);
  PEdge *f = find_left(v, sweep);
  check_diagonal(v, f->helper, diagonals);
  f->helper = v;
}

void handle_split (PVertex *v, PEdgeSet &sweep, PVertices &diagonals)
{
  PEdge *e = v->edge, *f = find_left(v, sweep);
  diagonals.push_back(v);
  diagonals.push_back(f->helper);
  f->helper = v;
  e->node = sweep.insert(e).first;
  e->in = true;
  e->helper = v;
}

PEdge * find_left (PVertex *v, PEdgeSet &sweep)
{
  PEdgeIter iter = sweep.upper_bound(v->edge);
  --iter;
  return *iter;
}

void check_diagonal (PVertex *v, PVertex *w, PVertices &diagonals)
{
  if (w->type() == Merge) {
    diagonals.push_back(v);
    diagonals.push_back(w);
  }
}

void delete_loop (PVertices &vertices)
{
  for (unsigned int i = 0; i < vertices.size(); ++i) {
    vector<PEdge *> edges;
    PEdge *e = vertices[i]->edge, *e0 = e;
    do {
      edges.push_back(e);
      e = e->next;
    }
    while (e != e0);
    for (unsigned int j = 0; j < edges.size(); ++j)
      delete edges[j];
    delete vertices[i];
  }
}

void triangulate_monotonic (PVertices &vertices, Triangles &triangles)
{
  for (unsigned int i = 0; i < vertices.size(); ++i) {
    PEdge *e0 = vertices[i]->edge, *e = e0;
    do {
      if (!e->flag) {
	Points pl, pr;
	e->monotonic_loop(pl, pr);
	triangulate_monotonic(pl, pr, triangles);
      }
      e = e->next;
    }
    while (e != e0);
  }
}

void triangulate_monotonic (const Points &pl, const Points &pr,
			    Triangles &triangles)
{
  bool rflag = true, nrflag;
  int nl = 0, nr = 1, lmax = pl.size(), rmax = pr.size();
  Points stack;
  stack.push_back(pr[0]);
  while (nl < lmax && nr < rmax) {
    nrflag = pr[nr].y > pl[nl].y;
    if (rflag && nrflag)
      triangulate_add(pr[nr++], stack, rflag, triangles);
    else if (rflag)
      triangulate_pop(pl[nl++], stack, triangles);
    else if (nrflag)
      triangulate_pop(pr[nr++], stack, triangles);
    else
      triangulate_add(pl[nl++], stack, rflag, triangles);
    rflag = nrflag;
  }
}

void triangulate_add (const Point &p, Points &stack, bool rflag,
		      Triangles &triangles)
{
  int n = stack.size() - 1;
  while (n > 0) {
    double k = signedArea(stack[n-1], stack[n], p);
    if (rflag ? k >= 0.0 : k <= 0.0)
      break;
    triangles.push_back(Triangle(p, stack[n], stack[n-1]));
    stack.pop_back();
    --n;
  }
  stack.push_back(p);
}

void triangulate_pop (const Point &p, Points &stack, Triangles &triangles)
{
  int n = stack.size() - 1;
  Point q = stack[n];
  while (n > 0) {
    triangles.push_back(Triangle(p, stack[n], stack[n-1]));
    --n;
  }
  stack.clear();
  stack.push_back(q);
  stack.push_back(p);
}

/*
main ()
{
  Profile reg;
  reg.push_back(Points());
  Points &poly = reg[0];
  read_points(poly, cin);
  Triangles tr;
  triangulate(reg, tr);
  write_triangles(tr, cout);
}
*/
