#ifndef POINT
#define POINT

#include <math.h>
#include <vector>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <string>

using namespace std;

class Point {
 public:
  Point () : x(0.0), y(0.0) {}
  Point (double ix, double iy) : x(ix), y(iy) {}
  bool operator== (const Point &b) const { return x == b.x && y == b.y; }
  bool operator!= (const Point &b) const { return x != b.x || y != b.y; }
  bool operator< (const Point &b) const { return x < b.x || x == b.x && y < b.y; }
  Point operator+ (const Point &b) const { return Point(x + b.x, y + b.y); }
  Point operator- (const Point &b) const { return Point(x - b.x, y - b.y); }
  Point operator- () const { return Point(- x, - y); }
  double operator% (const Point &b) const { return x*b.x + y*b.y; }
  double operator* (const Point &b) const { return x*b.y - y*b.x; }
  double dist2 (const Point &b) const 
    { return (x - b.x)*(x - b.x) + (y - b.y)*(y - b.y); }
  double distance (const Point &b) const { return sqrt(dist2(b)); }
  void polar (double &r, double &a) const;
  double norm () const {return sqrt(x*x + y*y); }
  void normalize () 
    { double d = norm(); if (d == 0.0) return; x /= d; y /= d; }

  double x, y;
};

typedef vector<Point> Points;

typedef vector<Points> Profile;

inline Point operator* (double k, const Point &p)
{
  return Point(k*p.x, k*p.y);
}

inline istream & operator>> (istream &in, Point &p)
{
  return in >> p.x >> p.y;
}

inline ostream & operator<< (ostream &out, Point &p)
{
  return out << p.x << " " << p.y;
}

inline double signedArea (const Point &a, const Point &b, const Point &c)
{
  return (c - b)*(a - b);
}

bool pointInTriangle (const Point &p, const Point &a, const Point &b,
		      const Point &c);

double signedArea (const Points &polygon);

Point random_point (double *bbox);

double random (double rmin, double rmax);

void read_points (Points &points, istream &istr);

void write_points (const Points &points, ostream &ostr);

void write_profile (const Profile &profile, ostream &ostr);

class Angle {
 public:
  Angle () {}
  Angle (double ic, double is, double d = -1.0)
    { if (d == -1.0) d = sqrt(ic*ic + is*is); c = ic/d; s = is/d; }
  Angle (const Point &p)
    { double d = p.norm(); c = p.x/d; s = p.y/d; }
  Angle (double th) : c(cos(th)), s(sin(th)) {}
  double angle () const { return atan2(s, c); }
  Angle inv () const { return Angle(- c, - s, 1.0); }
  bool operator== (const Angle &b) const { return s == b.s && c == b.c; }
  bool operator!= (const Angle &b) const { return s != b.s || c != b.c; }
  Angle operator+ (const Angle &b) const
    { return Angle(c*b.c - s*b.s, s*b.c + c*b.s); }
  Angle operator- () const
    { return Angle(c, - s); }
  double operator % (const Angle &b) const { return c*b.c + s*b.s; }
  double operator * (const Angle &b) const { return c*b.s - s*b.c; }
  Point operator * (const Point &p) const 
    { return Point(c*p.x - s*p.y, s*p.x + c*p.y); }
  bool in_arc (const Angle &s, const Angle &e) const;
  bool in_arc_strict (const Angle &s, const Angle &e) const;
  Angle midpoint (const Angle &e) const;

  double c, s;
};

inline Point operator* (double k, const Angle &a)
{
  return Point(k*a.c, k*a.s);
}

typedef vector<Angle> Angles;

class Triangle {
 public:
  Triangle () {}
  Triangle (const Point &ia, const Point &ib, const Point &ic)
    : a(ia), b(ib), c(ic) {}

  Point a, b, c;
};

typedef vector<Triangle> Triangles;

void write_triangles (const Triangles &triangles, ostream &ostr);

#endif
