#include "postscript.h"

PsView::PsView (int iw, int ih, double *bbox, ostream *iostr) 
  : w(iw), h(ih), ostr(iostr)
{
  xs = w/(bbox[1] - bbox[0]);
  xb = - xs*bbox[0];
  ys = h/(bbox[3] - bbox[2]);
  yb = - ys*bbox[2];
}

Point PsView::scale (const Point &p) const
{
  return Point(xs*p.x + xb, ys*p.y + yb);
}

void PsView::open () const
{
  *ostr << "%! catl output" << endl
	<< "%%BoundingBox: 0 0 " << w << " " << h << endl 
	<< "%%EndComments" << endl << endl
	<< "newpath 0 0 moveto " << w << " 0 lineto " << w << " " 
	<< h << " lineto 0 " << h << " lineto closepath clip" << endl;
}

void PsView::close () const
{
  *ostr << "showpage" << endl;
}

void PsView::polygon (const Points &points, double width, float *color, 
		      bool close, bool fill) const
{
  if (points.empty()) return;
  Point p0 = scale(points[0]);
  *ostr << "newpath " << p0.x << " " << p0.y << " moveto ";
  for (unsigned int i = 1; i < points.size(); ++i) {
    Point p = scale(points[i]);
    *ostr << p.x << " " << p.y << " lineto ";
  }
  if (fill)
    setprops(width, color, "closepath fill");
  else if (close)
    setprops(width, color, "closepath stroke");
  else
    setprops(width, color, "stroke");
}

void PsView::arrow (double x1, double y1, double x2, double y2,
		    double width, float *color, int length) const
{
  Point t = scale(Point(x1, y1)), h = scale(Point(x2, y2));
  double angle = atan2(y2 - y1, x2 - x1);
  double s = sin(angle), c = cos(angle);
  double ax1 = - length*(s + c), ax2 = length*(s - c),
    ay1 = - ax2, ay2 = ax1;
  *ostr << "newpath " << t.x << " " << t.y << " moveto "
	<< h.x << " " << h.y << " lineto "
	<< ax1 << " " << ay1 << " rlineto " << h.x << " " << h.y
	<< " moveto " << ax2 << " " << ay2 << " rlineto 1 setlinejoin ";
  setprops(width, color, "stroke");
}

void PsView::bullet (const Point &p, int r, float *color) const
{
  Point sp = scale(p);
  *ostr << "newpath " << sp.x << " " << sp.y << " " << r << " 0 360 arc ";
  setprops(1, color, "fill");
}

void PsView::setprops (double width, float *color, string action) const
{
  *ostr << width << " setlinewidth " 
	<< color[0] << " " << color[1] << " " << color[2] << " setrgbcolor "
	<< action << endl;
}

void PsView::draw_system (System *s, double *c, double width, bool fill, 
			  double acc) const
{
  open();
  for (unsigned int i = 0; i < s->parts.size(); ++i) {
    double cp[3];
    Part *part = s->parts[i];
    part->config(c[i], cp);
    draw_part(part, cp, width, fill, acc);
  }
  close();
}

void PsView::draw_part (Part *part, double *c, double width, bool fill, 
			double acc) const
{
  Part *mp = part->move(Angle(c[2]), Point(c[0], c[1]));
  double z[2];
  mp->form_profiles(acc, z);
  if (fill)
    for (unsigned int i = 0; i < mp->profiles.size(); ++i)
      draw_profile_filled(mp->profiles[i], width, part->color);
  else
    for (unsigned int i = 0; i < mp->profiles.size(); ++i)
      draw_profile(mp->profiles[i], width, part->color);
  delete mp;
}

void PsView::draw_profile_filled (const Profile &profile, double width,
				  float *color) const
{
  Triangles triangles;
  triangulate(profile, triangles);
  draw_triangles(triangles, color);
  float black[3] = {0, 0, 0};
  draw_profile(profile, width, black);
}

void PsView::draw_profile (const Profile &profile, double width, 
			   float *color) const
{
  for (unsigned int i = 0; i < profile.size(); ++i)
    polygon(profile[i], width, color, true, false);
}

void PsView::draw_triangles (const Triangles &triangles, float *color) const
{
  for (unsigned int i = 0; i < triangles.size(); ++i) {
    const Triangle &tr = triangles[i];
    Points pts;
    pts.push_back(tr.a);
    pts.push_back(tr.b);
    pts.push_back(tr.c);
    polygon(pts, 1, color, true, true);
  }
}

void PsView::draw_cspace (System *sys, Pair *pair, double *c, Trajectory *traj,
			  double *bbox, bool czone) const
{
  float black[3] = {0.0f, 0.0f, 0.0f}, grey[3] = {0.85f, 0.85f, 0.85f};
  float red[3] = {1.0f, 0.0f, 0.0f}, green[3] = {0.0f, 1.0f, 0.0f};
  float grey2[3] = {0.65f, 0.65f, 0.65f};
  Regions *blocked = complement(pair->regions, pair->bbox);
  Triangles triangles;
  double w = czone ? 0.5 : 1.0;
  for (unsigned int i = 0; i < blocked->size(); ++i)
    triangulate(blocked->at(i)->boundary, triangles);
  open();
  draw_triangles(triangles, grey);
  if (czone) {
    draw_triangles(pair->ctriangles, grey2);
    draw_regions(pair->czone, w, green);
  }
  draw_regions(blocked, w, black);
  draw_axes(bbox, 1, black);
  const Parts &pp = sys->parts;
  int n = pp.size(), ia = find(pp.begin(), pp.end(), pair->a) - pp.begin(),
    ib = find(pp.begin(), pp.end(), pair->b) - pp.begin();
  draw_trajectory(traj, n, ia, ib, 1, red);
  Point p(c[ia], c[ib]);
  bullet(p, 2, green);
  close();
  delete_regions(blocked);
}

void PsView::draw_regions (Regions *regions, double width, float *color) const
{
  if (!regions) return;
  for (unsigned int i = 0; i < regions->size(); ++i) {
    Profile &prof = regions->at(i)->boundary;
    if (prof[0].size() > 2)
      draw_profile(prof, width, color);
  }
}

void PsView::draw_axes (double *bbox, double width, float *color) const
{
  double x = 0.0, y = 0.0;
  if (bbox[0] > 0.0 || bbox[1] < 0.0)
    x = 0.5*(bbox[0] + bbox[1]);
  if (bbox[2] > 0.0 || bbox[3] < 0.0)
    y = 0.5*(bbox[2] + bbox[3]);
  arrow(bbox[0], y, bbox[1], y, width, color);
  arrow(x, bbox[2], x, bbox[3], width, color);
}

void PsView::draw_trajectory (Trajectory *traj, int np, int ia, int ib,
			      double width, float *color) const
{
  int n = traj->size()/(np+1);
  if (n < 2) return;
  Points pts;
  double x = 0.0, y = 0.0;
  for (int i = 0; i < n; ++i) {
    double nx = (*traj)[i*(np+1)+ia+1], ny = (*traj)[i*(np+1)+ib+1];
    if ((fabs(x) == M_PI || fabs(nx) == M_PI) && fabs(x - nx) > 6.28 ||
	(fabs(y) == M_PI || fabs(ny) == M_PI) && fabs(y - ny) > 6.28) {
      polygon(pts, width, color, false, false);
      pts.clear();
    }
    pts.push_back(Point(nx, ny));
    x = nx;
    y = ny;
  }
  polygon(pts, width, color, false, false);
}
