#include "region.h"

double Region::area () const
{
  double a = 0.0;
  for (unsigned int i = 0; i < boundary.size(); ++i)
    a += signedArea(boundary[i]);
  return a;
}

Point Region::interior_point () const
{
  const Points &b0 = boundary[0];
  const Point &a = b0[0], &b = b0[1], &c = *(b0.end()-2);
  if (boundary[0].size() == 2)
    return 0.5*(a + b);
  Angle ba = b - a, ca = c - a, u = ba.midpoint(ca);
  RegIntersections ri;
  ray_intersections(a, Point(u.c, u.s), ri);
  double w = 0.0, t = 0.0;
  for (unsigned int i = 0; i < ri.size(); ++i)
    if (!ri[i].in) {
      double tp = i == 0 ? 0.0 : ri[i-1].t;
      if (ri[i].t - tp > w) {
	w = ri[i].t - tp;
	t = 0.5*(ri[i].t + tp);
      }
    }
  return a + t*u;
}

bool Region::contains (const Point &p) const
{
  if (boundary.empty())
    return true;
  RegIntersections fi;
  Point v(0, 1);
  ray_intersections(p, v, fi);
  int n = fi.size();
  return boundary[0].empty() ? n%2 == 0 : n%2 == 1 && fi[n-1].nb == 0;
}

double Region::distance (const Point &p, int *edata, Point &q) const
{
  double d = point_boundary_distance(p, boundary[0], edata[1], q);
  edata[0] = 0;
  for (unsigned int i = 1; i < boundary.size() && d > 0.0; ++i) {
    int ei;
    Point qi;
    double di = point_boundary_distance(p, boundary[i], ei, qi);
    if (di < d) {
      d = di;
      edata[0] = i;
      edata[1] = ei;
      q = qi;
    }
  }
  return d;
}

void Region::ray_intersections (const Point &p, const Point &v,
				RegIntersections &res) const
{
  for (unsigned int i = 0; i < boundary.size(); ++i) {
    int np = res.size();
    ray_boundary_intersections(p, v, boundary[i], res);
    for (unsigned int j = np; j < res.size(); ++j)
      res[j].nb = i;
  }
  sort(res.begin(), res.end());
}

bool Region::ray_intersection (const Point &p, const Point &v,
			       double &t, Point &q, int *edata) const
{
  RegIntersections fi;
  ray_intersections(p, v, fi);
  unsigned int i = 0;
  if (!fi.empty() && fi[0].in && fi[0].nb == edata[0] &&
      fi[0].np == edata[1])
    i = 1;
  if (!(i < fi.size() && !fi[i].in))
    return false;
  t = fi[i].t;
  q = fi[i].p;
  edata[0] = fi[i].nb;
  edata[1] = fi[i].np;
  return true;
}

bool Region::path (const Point &p, const Point &q, Points &res) const
{
  RegIntersections fi;
  ray_intersections(p, q - p, fi);
  unsigned int i = 0, n = fi.size();
  while (i < n && fi[i].t <= 0.0)
    ++i;
  if (i < n && fi[i].in)
    ++i;
  unsigned int j = i;
  while (j < n && fi[j].t <= 1.0)
    ++j;
  res.push_back(p);
  while (i + 1 < j) {
    res.push_back(fi[i].p);
    if (fi[i].nb != fi[i+1].nb)
      return false;
    const Points &b = boundary[fi[i].nb];
    int np = b.size(), n1 = fi[i].np, n2 = fi[i+1].np, sgn, dk;
    sgn = n2 > n1 ? 1 : -1;
    dk = abs(n2 - n1) < np/2 ? sgn : - sgn;
    int k = n1 + dk;
    while (true) {
      if (k < 0)
	k = np - 1;
      else if (k == np)
	k = 0;
      if (k == n2)
	break;
      res.push_back(b[k]);
      k += dk;
    }
    res.push_back(fi[i+1].p);
    i += 2;
  }
  if (i + 1 == j)
    res.push_back(fi[i].p);
  if (contains(q)) {
    res.push_back(q);
    return true;
  }
  return false;
}

void delete_regions (Regions *regions)
{
  if (!regions) return;
  for (unsigned int i = 0; i < regions->size(); ++i)
    delete regions->at(i);
  delete regions;
}

double point_boundary_distance (const Point &p, const Points &boundary, 
				int &ei, Point &q)
{
  int j = boundary.size() - 1;
  double d = 1e20;
  for (unsigned int i = 0; i < boundary.size(); ++i) {
    Point qi;
    double di = point_segment_distance(p, boundary[j], boundary[i], qi);
    if (di < d) {
      d = di;
      ei = j;
      q = qi;
    }
    j = i;
  }
  return d;
}

double point_segment_distance (const Point &p, const Point &a, const Point &b,
			       Point &q)
{
  Point u = b - a, v = p - a;
  double uu = u%u, uv = u%v;
  if (uu == 0.0) {
    q = a;
    return p.dist2(a);
  }
  else if (uv >= 0.0 && uv <= uu) {
    q = a + (uv/uu)*u;
    double k = u*v;
    return k*k/uu;
  }
  else if (uv < 0.0) {
    q = a;
    return p.dist2(a);
  }
  else {
    q = b;
    return p.dist2(b);
  }
}

void ray_boundary_intersections (const Point &p, const Point &v, 
				 const Points &boundary, RegIntersections &res)
{
  int n = boundary.size();
  if (n < 2) return;
  int i = n - 1, j = 0;
  double si = (boundary[i] - p)*v;
  while (j < n) {
    double sj = (boundary[j] - p)*v, sk = sj;
    int k = j;
    while (si != 0.0 && sk == 0.0) {
      k = (k + 1)%n;
      sk = (boundary[k] - p)*v;
    }
    Point q;
    double t;
    bool inflag;
    if (si*sk < 0.0 &&
	ray_line_intersection(p, v, boundary[i], boundary[j], t, q, inflag))
      res.push_back(RegIntersection(t, q, inflag, i));
    if (k < j) return;
    i = k;
    si = sk;
    j = i + 1;
  }
}

bool ray_line_intersection (const Point &p, const Point &v, const Point &a,
			    const Point &b, double &t, Point &q, bool &inflag)
{
  Point u = b - a;
  double den = u*v;
  if (den == 0.0)
    return false;
  inflag = den > 0.0;
  if (p == b)
    q = b;
  else {
    double k = (p*v - a*v)/den;
    q = a + k*u;
  }
  t = fabs(v.x) < fabs(v.y) ? (q.y - p.y)/v.y : (q.x - p.x)/v.x;
  return t > 0.0 || t == 0.0 && inflag;
}
