#ifndef SYSTEM
#define SYSTEM

#include "mtriangulate.h"
#include "region.h"
#include "cs.h"

#define MINAREA 1e-10

typedef vector<Regions *> RegionsList;

typedef vector<Boundary *> Boundaries;

class glRegion {
 public:
  glRegion () : nvt(0), vb(0), vt(0) {}
  glRegion (Profile &profile, double z1, double z2);
  ~glRegion () { delete [] vb; delete [] vt; }

  vector<int> nvb;
  int nvt;
  double *vb, *vt;
};

typedef vector<glRegion *> glRegions;

void delete_glRegions (glRegions &regions);

class Slice {
 public:
  Slice (double ilb, double iub) : lb(ilb), ub(iub) {}
  ~Slice ();

  double lb, ub;
  Boundaries boundaries;
};

typedef vector<Slice *> Slices;

void delete_boundaries (Boundaries &boundaries);

typedef vector<Profile> Profiles;

class Part {
 public:
  Part (bool irot = false) : rot(irot) {}
  Part (string iname, float *icolor, bool irot, const Point &io, 
	const Angle &it, const Profile &loops);
  ~Part ();
  Regions * cspace (Part *b, double acc, double *bbox, double *ibbox = 0,
		    bool usebbox = false) const;
  void form_bbox (Part *b, double *bbox, double *ibbox) const;
  void form_bbox_rt (Part *b, double *bbox) const;
  void form_bbox_tt (Part *b, double *bbox) const;
  double rot_radius () const;
  void x_range (double &lb, double &ub) const;
  void y_range (double &lb, double &ub) const;
  Regions * cspace_slice (const Boundaries &ba, const Boundaries &bb,
			  Part *b, double acc, double *bbox) const;
  void form_regions (Arrangement *arr, Part *b, Boundary *ba, bool aout, 
		     Boundary *bb, bool bout, double acc, 
		     Regions *regions) const;
  bool overlap (Part *b, double u, double v, Boundary *ba, bool aout,
		Boundary *bb, bool bout) const;
  void read (istream &istr);
  Slice * read_slice (istream &istr, const Angle &th);
  Boundary * read_boundary (istream &istr, const Angle &th, bool flag) const;
  void read_sgroup (istream &istr, const Angle &th, Boundary &sgroup) const;
  Feature * shift_feature (Feature *f, double u) const;
  Part * move (const Angle &a, const Point &d) const;
  Part * offset (double d) const;
  bool wrap ();
  void config (double mp, double *c) const;
  void form_profiles (double acc, double *z);
  void triangulate ();
  void triangulate_3d (ostream &ostr) const;

  string name;
  float color[3];
  bool rot;
  Point o;
  Angle t;
  Slices slices;
  double x, v;
  Profiles profiles;
  glRegions regions;
  bool traversed;
};

typedef vector<Part *> Parts;

Boundary * move (Boundary *b, const Angle &a, const Point &d);

bool overlap (Boundary *b1, Boundary *b2);

bool contains (const Boundary &boundary, Point p);

void triangulate_profile_3d (const Profile &profile, double lb, double ub,
			     vector<double> &pts);

void triangulate_profile_tb (const Triangles &triangles, double z, bool ccw,
			     vector<double> &pts);

void triangulate_profile_s (const Profile &profile, double lb, double ub,
			    vector<double> &pts);

void add_point_3d (double x, double y, double z, vector<double> &pts);

void write_triangles_vtk (ostream &ostr, vector<double> &pts);

void triangulate_fa (Part *a, Part *b, double acc, double *bbox,
		     int axis, double *aint, vector<double> &pts);

void extrude (const Triangles &triangles, int axis, double *aint,
	      vector<double> &pts);

void transpose (const vector<double> &apts, int axis, vector<double> &pts);

void triangulate_patches (const Boundary &a, const Boundary &b,
			  double dt, double acc, ostream &ostr);

void triangulate_patch (Feature *f, Feature *g, double dt, double acc,
			int k, ostream &ostr);

void triangulate_patch (Feature *f, Feature *g, double dt, double acc,
			int k, vector<double> &pts);

void triangulate_patch_aa (Feature *arc1, Feature *arc2, double dt, 
			   double acc, int k, vector<double> &pts);

void triangulate_patch_aa_aux (Feature *arc1, Feature *arc2, double dt, 
			       double acc, int k, vector<double> &pts);

void triangulate_strip (Points &lb, double lth, Points &ub, double uth,
			int k, vector<double> &pts);

void resample (const Points &points, int n, Points &res);

void triangulate_patch_al (Feature *arc, Feature *line, double dt,
			   int k, vector<double> &pts);

void triangulate_patch_la (Feature *line, Feature *arc, double dt,
			   int k, vector<double> &pts);

void sampled_cspace (Part *a, Part *b, double ths, double the,
		     double dt, double acc, int k, ostream &ostr);

void sliced_cspace (Part *a, Part *b, double ths, double the,
		    double dt, double acc, ostream &ostr);

void reflect_points (const Points &pts, Points &rpts);

void reflect_profile (const Profile &prof, Profile &rprof);

Feature * read_feature (istream &istr, const Angle &th);

Regions * intersection (const RegionsList &rlist, double *bbox);

void region_segments (Region *region, double *bbox, Segments &segs);

bool bbox_segment (const Point &t, const Point &h, double *bbox);

Regions * complement (Regions *regions, double *bbox);

class ContactElt {
 public:
  ContactElt () {}
  ContactElt(double ix, Contact *ic) : x(ix), c(ic) {}
  bool operator< (const ContactElt &b) const { return x < b.x; }

  double x;
  Contact *c;
};

class MotionElt {
 public:
  MotionElt () : part(0) {}
  MotionElt (Part *ipart, double iv) : part(ipart), v(iv) {}

  Part *part;
  double v;
};

typedef vector<MotionElt> MotionElts;

class Motion {
 public:
  Motion (double itime = 0) : time(itime) {}

  double time;
  MotionElts elts;
};

typedef vector<Motion> Motions;

class Pair;

class SimulateEvent {
 public:
  SimulateEvent () : pair(0) {}

  Pair *pair;
  double time;
  int edge[2];
  Point p;
};

class Pair {
 public:
  Pair (Part *ia, Part *ib, Part *ira = 0, Part *itb = 0, double iacc = 0.0) 
    : a(ia), b(ib), ra(ira), tb(itb), acc(iacc), regions(0), czone(0),
    region(0) {}
  ~Pair ();
  void set_region ();
  bool set_velocity (bool aflag, bool &blocked);
  bool next_event (SimulateEvent &event);
  void wrap ();
  void contact_segment (Point &t, Point &h) const;
  void form_czone (double d);
  bool path (const Point &p, const Point &q, Points &res,
	     bool optimize = false) const;
  Region * enclosing_region (const Point &p) const;

  Part *a, *b, *ra, *tb;
  double bbox[4], acc;
  Regions *regions, *czone;
  Region *region;
  int edge[2];
  bool free, traversed;
  Triangles ctriangles;
  glRegions glregions, glczone;
  glRegion gltraj;
};

typedef vector<Pair *> Pairs;

Pair * form_pair (Part *a, Part *b, double acc, double *bbox = 0,
		  bool usebbox = false);

void triangulate_regions (Regions *regions, Triangles &triangles);

typedef vector<double> Trajectory;

class System {
 public:
  System () {}
  ~System ();
  void read (istream &istr);
  void read_motions (istream &istr, Motions &motions, double *c0,
		     double *pb) const;
  void simulate (double *c0, const Motions &motions, Trajectory &traj);
  void set_state (double *state);
  void write_state (double t, Trajectory &traj) const;
  void set_regions ();
  bool set_velocity (const Motion &motion);
  bool set_velocity (Part *p);
  bool next_event (SimulateEvent &event) const;
  void update_state (const SimulateEvent &event);
  void update_state (double time);
  bool wrap ();
  void form_profiles (double *z);
  void form_pair_trajectories (Trajectory &traj);
  void form_czone (double d);

  Parts parts;
  Pairs pairs;
};

#endif
